from .handler import Handler
from arjuna.setuext.guiauto.impl.base.conditions import *

class GuiAutomatorConditions(Handler):

    def __init__(self, automator):
        super().__init__(automator)

    def AlertIsPresent(self):
        caller = DynamicCaller(self.automator.alert_handler.is_alert_present)
        return BooleanCondition(caller, True)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import types
from arjuna.lib.value import AnyRefValue


class DataRecord:
    def __init__(self, *vargs, **kwargs):
        self.__indexed = tuple([AnyRefValue(v) for v in vargs])
        self.__named = types.MappingProxyType({i.lower(): AnyRefValue(j) for i, j in kwargs.items()})

    def value_at(self, i):
        return self.__indexed[i]

    def value_named(self, name):
        return self.__named[name.lower()]

    def indexed_values(self):
        return self.__indexed

    def named_values(self):
        return self.__named

    def __str__(self):
        parts = ["Indexed"]
        if not self.indexed_values():
            parts.append("<empty>")
        else:
            for i,v in enumerate(self.indexed_values()):
                parts.append("[{}] {}".format(i, v.as_string()))
        parts.append("Named")
        if not self.named_values():
            parts.append("<empty>")
        else:
            for i,v in self.named_values().items():
                parts.append("[{}] {}".format(i, v.as_string()))
        return "\n".join(parts)

    def is_empty(self):
        return not self.indexed_values() and not self.named_values()

    def is_list_empty(self):
        return not self.indexed_values()

    def is_map_empty(self):
        return not self.named_values()

    def has_key(self, key):
        return key.lower() in self.__named

    def string(self, key):
        return self.value_named(key).as_string()

    def has_index(self, index):
        return len(self.__indexed) > index



from arjuna.lib.value import AbstractValueMap, StringKeyValueMap
from arjuna.client.core.connector import BaseSetuObject, SetuArg
from arjuna.tpi.enums import ArjunaOption
from arjuna.tpi.enums import GuiAutomationContext, BrowserName, GuiAutomatorName
from arjuna.lib.value import AnyRefValue


class DefaultTestConfig(BaseSetuObject):

    def __init__(self, test_session, name, setu_id, arjuna_options, user_options):
        super().__init__()
        self.__session = test_session
        self.__name = name

        self._set_setu_id(setu_id)
        self._set_self_setu_id_arg("configSetuId")
        self._set_test_session_setu_id_arg(self.__session.get_setu_id())
        self.arjuna_options = {DefaultTestConfig.normalize_arjuna_option_str(k): v for k,v in arjuna_options.items()}
        self.user_options = user_options

    def get_test_session(self):
        return self.__session

    def __fetch_config_option_value(self, setu_action_type, option_str):
        response = self._send_request(setu_action_type, SetuArg.arg("option", option_str))
        return response.get_value()

    @staticmethod
    def normalize_option_str(option_str):
        return option_str.upper().strip().replace(".", "_")

    @staticmethod
    def normalize_arjuna_option_str(option_str):
        return ArjunaOption[DefaultTestConfig.normalize_option_str(option_str)]

    def get_arjuna_option_value(self, option):
        arjuna_option = option
        if type(option) is str:
            arjuna_option = DefaultTestConfig.normalize_arjuna_option_str(option)
        return AnyRefValue(self.arjuna_options[arjuna_option])

    def get_user_option_value(self, option):
        user_option = DefaultTestConfig.normalize_option_str(option)
        return AnyRefValue(self.user_options[user_option])

    def get_name(self):
        return self.__name

    def get_gui_auto_context(self):
        return self.get_arjuna_option_value(ArjunaOption.GUIAUTO_CONTEXT).as_enum(GuiAutomationContext)

    def get_browser_type(self):
        return self.get_arjuna_option_value(ArjunaOption.BROWSER_NAME).as_enum(BrowserName)

    def get_browser_version(self):
        return self.get_arjuna_option_value(ArjunaOption.BROWSER_VERSION).as_string()

    def get_browser_binary_path(self):
        return self.get_arjuna_option_value(ArjunaOption.BROWSER_BIN_PATH).as_string()

    def get_test_run_env_name(self):
        return self.get_arjuna_option_value(ArjunaOption.TESTRUN_ENVIRONMENT).as_string()

    def get_screenshots_dir(self):
        return self.get_arjuna_option_value(ArjunaOption.SCREENSHOTS_DIR).as_string()

    def get_log_dir(self):
        return self.get_arjuna_option_value(ArjunaOption.LOG_DIR).as_string()

    def get_gui_auto_max_wait_time(self):
        return self.get_arjuna_option_value(ArjunaOption.GUIAUTO_MAX_WAIT).as_int()


class ArjunaOptionContainer(AbstractValueMap):

    def __init__(self, object_map=None, headers=None, objects=None):
        super().__init__(object_map, headers, objects)

    def _format_key_as_str(self, key):
        return str(key)


class ConfigContainer:

    def __init__(self):
        self.__arjuna_options = ArjunaOptionContainer()
        self.__user_options = StringKeyValueMap()

    @property
    def arjuna_options(self):
        return self.__arjuna_options

    @property
    def user_options(self):
        return self.__user_options

    def set_arjuna_option(self, arjuna_option, obj):
        if isinstance(arjuna_option, ArjunaOption):
            self.__arjuna_options.add_object(arjuna_option.name, obj)
        else:
            normalized_option = DefaultTestConfig.normalize_option_str(arjuna_option)
            self.__arjuna_options.add_object(ArjunaOption[normalized_option], obj)
        return self

    def set_user_option(self, option, obj):
        self.__user_options.add_object(DefaultTestConfig.normalize_option_str(option), obj)
        return self

    def set_option(self, option, obj):
        normalized_option = DefaultTestConfig.normalize_option_str(option)
        try:
            arj_option = ArjunaOption[normalized_option]
            self.set_arjuna_option(arj_option, obj)
        except:
            self.set_user_option(option, obj)
        return self

    def add_options(self, options):
        for option, obj in options.items():
            self.set_option(option, obj)
        return self

    def is_empty(self):
        if not self.arjuna_options and not self.user_options:
            return True
        return False


class _ConfigBuilder:

    def __init__(self, test_session, config_map, conf_trace, code_mode):
        self.__code_mode = code_mode
        self.__test_session = test_session
        self.__config_container = ConfigContainer()

        self.__config_map = config_map
        if "default_config" in config_map:
            self.__parent_config = config_map["default_config"]
        else:
            self.__parent_config = None

        # For Unitee
        self.__conf_trace = conf_trace

    def parent_config(self, config):
        self.__parent_config = config

    def arjuna_option(self, option, obj):
        self.__config_container.set_arjuna_option(option, obj)
        return self

    def user_option(self, option, obj):
        self.__config_container.set_user_option(option, obj)
        return self

    def options(self, option_map):
        self.__config_container.set_options(option_map)
        return self

    def selenium(self):
        self.__config_container.set_arjuna_option(ArjunaOption.GUIAUTO_AUTOMATOR_NAME, GuiAutomatorName.SELENIUM.name)
        return self

    def appium(self, context):
        self.arjuna_option(ArjunaOption.GUIAUTO_AUTOMATOR_NAME, GuiAutomatorName.APPIUM.name)
        self.arjuna_option(ArjunaOption.GUIAUTO_CONTEXT, context)
        return self

    def chrome(self):
        self.arjuna_option(ArjunaOption.BROWSER_NAME, BrowserName.CHROME.name)
        return self

    def firefox(self):
        self.arjuna_option(ArjunaOption.BROWSER_NAME, BrowserName.FIREFOX.name)
        return self

    def headless_mode(self):
        self.arjuna_option(ArjunaOption.BROWSER_HEADLESS_MODE, True)
        return self

    def guiauto_max_wait_time(self, seconds):
        self.arjuna_option(ArjunaOption.GUIAUTO_MAX_WAIT, seconds)
        return self

    def app(self, path):
        self.arjuna_option(ArjunaOption.MOBILE_APP_FILE_PATH, path)
        return self

    def mobile_device_name(self, name):
        self.arjuna_option(ArjunaOption.MOBILE_DEVICE_NAME, name)
        return self

    def mobile_device_udid(self, udid):
        self.arjuna_option(ArjunaOption.MOBILE_DEVICE_UDID, udid)
        return self

    def build(self, config_name="default_config"):
        if not self.__config_container.arjuna_options.str_items() or not self.__config_container.user_options.items():
            if not self.__parent_config:
                if config_name != "default_config":
                    self.__config_map[config_name] = self.__config_map["default_config"]
            else:
                self.__config_map[config_name] = self.__parent_config
            return

        if not self.__parent_config:
            config = self.__test_session.register_config(
                config_name,
                self.__config_container.arjuna_options.str_items(),
                self.__config_container.user_options.items())
        else:
            config = self.__test_session.register_child_config(
                config_name,
                self.__parent_config.get_setu_id(),
                self.__config_container.arjuna_options.str_items(),
                self.__config_container.user_options.items()
            )

        self.__config_map[config_name] = config
        if self.__code_mode:
            if config_name not in self.__conf_trace:
                self.__conf_trace[config_name] = {"arjuna_options": set(), "user_options" : set()}
            self.__conf_trace[config_name]["arjuna_options"].update(self.__config_container.arjuna_options.items().keys())
            self.__conf_trace[config_name]["user_options"].update(self.__config_container.user_options.items().keys())


class DefaultTestContext:

    def __init__(self, test_session, name, parent_config=None):
        self.__test_session = test_session
        self.__name = name
        self.__parent_config_setu_id = parent_config and parent_config.get_setu_id() or None
        self.__configs = dict()
        self.__conf_trace = dict()

    def ConfigBuilder(self, code_mode=True):
        return _ConfigBuilder(self.__test_session, self.__configs, self.__conf_trace, code_mode)

    def update_with_file_config_container(self, container):
        for config_name, conf in self.__configs.items():
            builder = self.ConfigBuilder(code_mode=False)
            builder.parent_config(conf)
            amap = container.arjuna_options
            umap = container.user_options
            if config_name in self.__conf_trace:
                if "arjuna_options" in self.__conf_trace[config_name]:
                    for k,v in container.arjuna_options.items():
                        if k not in self.__conf_trace[config_name]["arjuna_options"]:
                            builder.arjuna_option(k, v)
                else:
                    for k, v in container.arjuna_options.items():
                        builder.arjuna_option(k, v)
                if "user_options" in self.__conf_trace[config_name]:
                    for k,v in container.user_options.items():
                        if k not in self.__conf_trace[config_name]["user_options"]:
                            builder.user_option(k, v)
                else:
                    for k, v in container.user_options.items():
                        builder.user_option(k, v)
            else:
                for k in amap.keys():
                    builder.arjuna_option(k, amap.object(k))
                for k in umap.keys():
                        builder.user_option(k, umap.object(k))
            builder.build(config_name=config_name)

    def get_config(self, config_name="default_config"):
        return self.__configs[config_name]

    def _add_configs(self, configs):
        self.__configs.update(configs)

    def _add_conf_trace(self, conf_trace):
        self.__conf_trace.update(conf_trace)

    def _get_conf_trace(self):
        return self.__conf_trace

    def _get_configs(self):
        return self.__configs

    def _get_test_session(self):
        return self.__test_session

    def get_name(self):
        return self.__name


# For Unitee
class InternalTestContext(DefaultTestContext):

    def __init__(self, test_session, name, parent_config=None):
        super().__init__(test_session, name, parent_config)

    def add_config(self, config):
        self.__configs[config.get_name()] = config

    def clone(self):
        out_context = InternalTestContext(self._get_test_session(), self.get_name())
        out_context._add_configs(self._get_configs())
        out_context._add_conf_trace(self._get_conf_trace())
        return out_context

    def clone_for_user(self):
        out_context = DefaultTestContext(self._get_test_session(), self.get_name())
        out_context._add_configs(self._get_configs())
        out_context._add_conf_trace(self._get_conf_trace())
        return out_context

    def update_from_context(self, context):
        self._add_configs(context._get_configs())


class CliArgsConfig:

    def __init__(self, arg_dict):
        self.__aco = {}
        self.__ato = {}
        self.__uco = {}
        self.__uto = {}

        kinds = {
            "aco": self.__aco,
            "ato": self.__ato,
            "uco": self.__uco,
            "uto": self.__uto
        }

        lower_actual_key_map = {i.lower():i for i in arg_dict}
        for kind in kinds:
            if kind in lower_actual_key_map:
                actual_key = lower_actual_key_map[kind]
                d_item = arg_dict[actual_key]
                if d_item:
                    for entry in d_item:
                        k,v = entry
                        kinds[kind][k.lower()] = v
                del arg_dict[actual_key]

        for akey, avalue in arg_dict.items():
            self.__aco[akey.lower()] = avalue

    def as_map(self):
        return {
            "arjunaCentralOptions": self.__aco,
            "arjunaTestOptions": self.__ato,
            "userCentralOptions": self.__uco,
            "userTestOptions": self.__uto
        }


import re
from enum import Enum, auto
from collections import namedtuple

class MobileNativeLocateWith(Enum):
    ID = auto()
    NAME = auto() 
    XPATH = auto() 
    CLASS_NAME = auto() 
    LINK_TEXT = auto() 
    PARTIAL_LINK_TEXT = auto() 
    TAG_NAME = auto()
    TEXT = auto() 
    TITLE = auto() 
    PARTIAL_TEXT = auto() 
    TYPE = auto() 
    VALUE = auto() 
    IMAGE_SRC = auto()

class MobileWebLocateWith(Enum):
    ID = auto()
    NAME = auto() 
    XPATH = auto() 
    CSS_SELECTOR = auto()
    CLASS_NAME = auto() 
    LINK_TEXT = auto() 
    PARTIAL_LINK_TEXT = auto() 
    TAG_NAME = auto()
    TEXT = auto() 
    TITLE = auto() 
    PARTIAL_TEXT = auto() 
    TYPE = auto() 
    VALUE = auto() 
    IMAGE_SRC = auto()

class NativeLocateWith(Enum):
    ID = auto()
    NAME = auto() 
    XPATH = auto() 
    CLASS_NAME = auto() 
    LINK_TEXT = auto() 
    PARTIAL_LINK_TEXT = auto() 
    TEXT = auto() 
    TITLE = auto() 
    PARTIAL_TEXT = auto() 
    TYPE = auto() 
    VALUE = auto() 
    IMAGE_SRC = auto()

class VisualLocateWith(Enum):
    IMAGE = auto()

class WebLocateWith(Enum):
    ID = auto()
    NAME = auto() 
    XPATH = auto() 
    CSS_SELECTOR = auto()
    CLASS_NAME = auto() 
    LINK_TEXT = auto() 
    PARTIAL_LINK_TEXT = auto() 
    TAG_NAME = auto()
    CLASS_NAMES = auto()
    TEXT = auto() 
    TITLE = auto() 
    PARTIAL_TEXT = auto() 
    ATTR_VALUE = auto()
    ATTR_PARTIAL_VALUE = auto()
    TYPE = auto() 
    VALUE = auto() 
    IMAGE_SRC = auto()

class GenericLocateWith(Enum):
    ID = auto()
    NAME = auto() 
    XPATH = auto() 
    CSS_SELECTOR = auto()
    CLASS_NAME = auto() 
    LINK_TEXT = auto() 
    PARTIAL_LINK_TEXT = auto() 
    TAG_NAME = auto()
    CLASS_NAMES = auto()
    TEXT = auto() 
    TITLE = auto() 
    PARTIAL_TEXT = auto() 
    TYPE = auto() 
    VALUE = auto() 
    ATTR_VALUE = auto()
    ATTR_PARTIAL_VALUE = auto()
    IMAGE_SRC = auto()
    IMAGE = auto()
    INDEX = auto()

Locator = namedtuple("Locator", ("ltype", "lvalue"))
GuiGenericLocator = namedtuple("GuiGenericLocator", ("ltype", "lvalue"))

class GuiElementType(Enum):
    TEXTBOX = auto()
    PASSWORD = auto()
    LINK = auto()
    BUTTON = auto()
    SUBMIT_BUTTON = auto()
    DROPDOWN = auto()
    CHECKBOX = auto()
    RADIO = auto()
    IMAGE = auto()

class GuiElementMetaData:
    XPATH_TWO_ARG_VALUE_PATTERN = r'^\s*\[\s*(\w+)\s*\]\s*\[\s*(\w+)\s*\]$'

    BASIC_LOCATORS = {
        GenericLocateWith.ID,
        GenericLocateWith.NAME,
        GenericLocateWith.CLASS_NAME,
        GenericLocateWith.LINK_TEXT,
        GenericLocateWith.PARTIAL_LINK_TEXT,
        GenericLocateWith.XPATH,
        GenericLocateWith.CSS_SELECTOR,
        GenericLocateWith.TAG_NAME,
        GenericLocateWith.IMAGE,
        GenericLocateWith.INDEX,
    }

    XTYPE_LOCATORS = {
        GuiElementType.TEXTBOX: "//input[@type='text']",
        GuiElementType.PASSWORD: "//input[@type='password']",
        GuiElementType.LINK: "//a",
        GuiElementType.BUTTON: "//input[@type='button']",
        GuiElementType.SUBMIT_BUTTON: "//input[@type='submit']",
        GuiElementType.DROPDOWN: "//select",
        GuiElementType.CHECKBOX: "//input[@type='checkbox']",
        GuiElementType.RADIO: "//input[@type='radio']",
        GuiElementType.IMAGE: "//img",
    }

    XPATH_LOCATORS = {
        GenericLocateWith.TEXT : "//*[text()='{}']",
        GenericLocateWith.PARTIAL_TEXT : "//*[contains(text(),'{}')]",
        GenericLocateWith.VALUE : "//*[@value='{}']",
        GenericLocateWith.TITLE : "//*[@title='{}']",
        GenericLocateWith.IMAGE_SRC : "//img[@src='{}']"
    }

    XPATH_TWO_ARG_LOCATORS = {
        GenericLocateWith.ATTR_VALUE : "//*[@{}='{}']",
        GenericLocateWith.ATTR_PARTIAL_VALUE : "//*[contains(@{},'{}')]"
    }

    def __init__(self, raw_locators):
        self.__raw_locators = raw_locators
        self.__locators = []
        self.__process()

    @property
    def locators(self):
        return self.__locators

    @property
    def raw_locators(self):
        return self.__raw_locators

    def __process(self):
        for raw_locator in self.__raw_locators:
            rltype = raw_locator.ltype
            rlvalue = raw_locator.lvalue
            try:
                generic_locate_with = GenericLocateWith[rltype.upper()]
            except:
                raise Exception("Invalid locator across all automators: {}".format(rltype))
            else:
                if generic_locate_with in self.BASIC_LOCATORS:
                    self.__add_locator(generic_locate_with, rlvalue)
                elif generic_locate_with in self.XPATH_LOCATORS:
                    self.__add_locator(GenericLocateWith.XPATH, self.XPATH_LOCATORS[generic_locate_with].format(rlvalue))
                elif generic_locate_with in self.XPATH_TWO_ARG_LOCATORS:
                    parts = None
                    try:
                        parts =  re.search(GuiElementMetaData.XPATH_TWO_ARG_VALUE_PATTERN, rlvalue).groups()
                        print(parts)
                    except:
                        raise Exception("Value {} for {} is misformatted. Attribute name and attribute value should be supplied as: [<arg>][<value>]".format(rlvalue, rltype))
                    self.__add_locator(GenericLocateWith.XPATH, self.XPATH_TWO_ARG_LOCATORS[generic_locate_with].format(parts[0], parts[1]))
                elif generic_locate_with == GenericLocateWith.TYPE:
                    try:
                        elem_type = GuiElementType[rlvalue.upper()]
                    except:
                        raise Exception("Unsupported element type for XTYPE locator: " + rlvalue)
                    else:
                        self.__add_locator(GenericLocateWith.XPATH, self.XTYPE_LOCATORS[elem_type])
                elif generic_locate_with == GenericLocateWith.CLASS_NAMES:
                    self.__add_locator(GenericLocateWith.CSS_SELECTOR, re.sub(r'\s+', '.', "." + rlvalue.replace('.', ' ').strip()))
                else:
                    raise Exception("Locator not supported yet by Setu: " + rltype)

    def __add_locator(self, locator_type, locator_value):
        self.locators.append(GuiGenericLocator(locator_type, locator_value))

    @classmethod
    def createEMD(self, locators):
        processed_locators = []
        for locator in locators:
            processed_locators.append(
                Locator(ltype=locator["withType"], lvalue=locator["withValue"])
            )
        return GuiElementMetaData(processed_locators)

class SimpleGuiElementMetaData(GuiElementMetaData):

    def __init__(self, locator_type, locator_value):
        super().__init__([Locator(ltype=locator_type, lvalue=locator_value)])



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

def split(in_string, delimiter="\t", uc=False):
    def upper(i):
        return uc and i.upper() or i
    return [upper(i.strip()) for i in in_string.split(delimiter)]



from flask import request
from flask_restful import Resource
from .objmgr import SetuSvcObjectManager
from arjuna.setu.testsession.adapter import TestSessionHandler
from arjuna.client.core.config import ArjunaComponent
from arjuna.client.core.action import *
from arjuna.setuext.guiauto.adapter.automator import GuiAutomatorHandler

from arjuna.setu import Setu
from arjuna.setu.adapter import Handler

import os

def create_success_response(data):
    res = {'result': 'success', 'responseData': data}, 200
    Setu.get_logger().debug("Setu Action Response: {}".format(res))
    return res


def create_error_response(emsg, etrace=None):
    etrace = etrace and etrace or "na"
    res = {'result': 'error', 'emessage': emsg, 'etrace': str(etrace)}, 500
    Setu.get_logger().debug("Setu Action Response (Error): {}".format(res))
    return res

component_action_map = {
    "SETU" : SetuActionType,
	"TEST_SESSION" : TestSessionActionType,
	"CONFIGURATOR" : ConfigActionType,
	"DATA_SOURCE" : DataSourceActionType,
	"GUI_AUTOMATOR" : GuiAutoActionType,
	"GUI" : GuiActionType
}

ACTION_HANDLER_MAP = {
    "SETU" : "_handle_setu_action",
    "TEST_SESSION" : "_handle_test_session_action",
    "CONFIGURATOR" : "_handle_configurator_action",
    "DATA_SOURCE" : "_handle_data_source_action",
    "GUI_AUTOMATOR" : "_handle_gui_automator_action",
    "GUI" : "_handle_gui_action"
}

class SetuSvc(Resource):

    def post(self):
        json_dict = request.get_json(force=True)
        Setu.get_logger().debug("Setu Action Request: {}".format(json_dict))
        json_component = json_dict["component"]
        json_action = json_dict["action"]
        action_type = None
        try:
            action_type = component_action_map[json_component][json_action.strip().upper()]
        except:
            return {'result': 'error', 'emessage': 'Invalid Setu action: Component: {} Action: {}'.format(json_component, json_action), 'etrace': 'NA'}, 500
        else:
            json_args = json_dict.get("args", {})
            return self.__call_handler(json_component, action_type, json_args)

    def __register_test_session(self, root_dir, cliConfig=None):
        handler = TestSessionHandler()

        config = handler.init(root_dir, cliConfig)
        SetuSvcObjectManager.register_testsession_handler(handler)
        out = {
            'testSessionSetuId': handler.setu_id,
            'configSetuId': config.setu_id,

        }

        out.update(config.as_json_dict())
        return out

    def __get_testsession_handler(self, component, action_type, json_args):
        try:
            tsession_id = json_args["testSessionSetuId"]
        except:
            return None, 'You must supply test session id in args for action {} for component {}. Your supplied JSON: {}'.format(action_type.name, component, json_args)
        else:
            handler = SetuSvcObjectManager.get_testsession_handler(tsession_id)
            del json_args["testSessionSetuId"]
            return handler, None

    def __call_handler(self, component, action_type, json_args):
        if component == "SETU":
            return self._handle_setu_action(action_type, json_args)
        elif component == "TEST_SESSION":
            return self._handle_test_session_action(action_type, json_args)
        else:
            testsession_handler, err = self.__get_testsession_handler(component, action_type, json_args)
            if not testsession_handler:
                return create_error_response(err)
            try:
                res = getattr(SetuSvc, ACTION_HANDLER_MAP[component])(
                    self,
                    testsession_handler,
                    action_type,
                    json_args
                )
            except Exception as f:
                import traceback
                ftrace = traceback.format_exc()
                return create_error_response(str(f), ftrace)
            else:
                return create_success_response(res)

    def _handle_setu_action(self, action_type, json_args):
        if action_type == SetuActionType.HELLO:
            return create_success_response({'value' : 'hello'})
        elif action_type == SetuActionType.EXIT:
            if not SetuSvcObjectManager.has_active_testsession():
                if os.name == 'nt':
                    os.system("taskkill /PID " + str(os.getpid()) + " /F")
                else :
                    os.system("kill " + str(os.getpid()))

    def _handle_test_session_action(self, action_type, json_args):
        if action_type == TestSessionActionType.INIT:
            try:
                root_dir = json_args["projectRootDir"]
                del json_args["projectRootDir"]
                res = self.__register_test_session(root_dir, **json_args)
                return create_success_response(res)
            except Exception as e:
                import traceback
                etrace = traceback.format_exc()
                return create_error_response(str(e), etrace)
        elif action_type == TestSessionActionType.END:
            pass

    def _handle_configurator_action(self, ts_handler, action_type, json_args):
        return ts_handler.conf_handler.take_action(action_type.name, json_args)

    def _handle_data_source_action(self, ts_handler, action_type, json_args):
        return ts_handler.datasource_handler.take_action(action_type, json_args)

    def _handle_gui_automator_action(self, ts_handler, action_type, json_args):
        if action_type == GuiAutoActionType.LAUNCH_AUTOMATOR:
            config = ts_handler.conf_handler.configurator.get_config(Handler.get_config_setuid(json_args))
            automator_handler = GuiAutomatorHandler(ts_handler.dispatcher)
            automator_handler.launch_automator(config, **json_args)
            ts_handler.register_gui_automator_handler(automator_handler)
            return {'automatorSetuId' : automator_handler.setu_id}
        elif action_type == GuiAutoActionType.QUIT_AUTOMATOR:
            handler = ts_handler.get_automator_handler(json_args)
            handler.quit_automator()
            ts_handler.deregister_gui_automator_handler(handler) 
        else:
            gui_setu_id = json_args.get("guiSetuId", None)
            handler = None
            if gui_setu_id is not None:
                handler = ts_handler.get_gui_handler(json_args)
            else:
                gui_automator_setu_id = json_args.get("automatorSetuId", None)
                if gui_automator_setu_id is None:
                    raise Exception("For gui component action either guiSetuId or automatorSetuId must be provided.")
                handler = ts_handler.get_automator_handler(json_args)
            return handler.take_action(action_type, json_args)

    def _handle_gui_action(self, ts_handler, action_type, json_args):
        automator_handler = ts_handler.get_automator_handler(json_args)
        return ts_handler.gui_manager.take_action(automator_handler, action_type, json_args)


'''
For consistency of Setu API across languages, all setu requester API follows Java coding convention in its interface.
This should also be adhered for any public classes and their public methods in a module internally used.

The structure of imports should be kept almost same so that users have little cognitive overhead.
'''


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os

from arjuna.lib.utils import data_utils
from arjuna.lib.types import constants


class FileReader:
    def __init__(self, file_path):
        self._fpath = file_path
        self._f = open(self._fpath, "r")

    def close(self):
        self._f.close()

    def read(self):
        return self._f.read()

    def _get_file(self):
        return self._f


class TextResourceReader(FileReader):
    def __init__(self, file_name):
        super().__init__(
            os.path.abspath(os.path.join(os.path.dirname(__file__), os.sep.join(["../..", "res", file_name]))))


class FileLineReader:
    def __init__(self, file_path):
        self.__filereader = open(file_path, "r")

    def __iter__(self):
        return self

    def next(self):
        l = self.__filereader.readline()
        if not l:
            self.close()
            raise StopIteration()
        elif l.strip().startswith('#'):
            return self.next()
        else:
            return l.rstrip()

    def close(self):
        self.__filereader.close()


class FileLine2ArrayReader:
    def __init__(self, file_path, delimiter="\t"):
        self.__filereader = FileLineReader(file_path)
        self.delimiter = delimiter
        self.headers = None
        self._populate_headers()

    def _populate_headers(self):
        try:
            self.headers = self.next()
        except:
            raise Exception("Invalid input file data. Empty headers line.")

    def next(self):
        line = self.__filereader.next()
        return data_utils.split(line.rstrip(), self.delimiter)

    def get_headers(self):
        return self.headers

    def close(self):
        self.__filereader.close()


class FileLine2MapReader:
    def __init__(self, file_path, delimiter="\t"):
        self.__filereader = FileLine2ArrayReader(file_path, delimiter)

    def get_headers(self):
        return self.__filereader.headers

    def close(self):
        self.__filereader.close()

    def next(self):
        line_parts = self.__filereader.next()

        if len(self.get_headers()) != len(line_parts):
            raise Exception(
                "Invalid input file data. Number of headers and data values do not match.{0}Headers:{1}{0}Data values:{2}{0}".format(
                    os.linesep, self.get_headers(), line_parts))
        else:
            return dict(zip(self.get_headers(), line_parts))



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.unitee.enums import *
from arjuna.lib.types.descriptors import *
class TestSourceDef:
    def __init__(self):
        self.pkg = None
        self.module = None
        self.func = None
        self.qname = None

        self.__tvars = None
        self.__threads = None
        self.__deps = None
        self.__drefs = None
        self.__dsources = []
        self.__instances = []

        self.__skipped = False
        self.__skip_code = None
        self.selected = False
        self.__unpicked = True
        self.__unpick_code = None

        self.consumed = False

        from arjuna.tpi import Arjuna
        self.console = Arjuna.get_console()
        self.logger = Arjuna.get_logger()
        self.central_config = Arjuna.get_central_config()

    @property
    def tvars(self):
        return self.__tvars

    @tvars.setter
    def tvars(self, tvars):
        self.__tvars = tvars

    @property
    def threads(self):
        return self.__threads

    @threads.setter
    def threads(self, threads):
        self.__threads = RangeBoundInt.validate(threads, 1, 100)

    @property
    def deps(self):
        return self.__deps

    @deps.setter
    def deps(self, value):
        self.__threads = value

    @property
    def drefs(self):
        return self.__drefs

    @drefs.setter
    def drefs(self, value):
        self.__threads = value

    @property
    def dsources(self):
        if self.__dsource is None:
            return DataDriver(dklass(DummyDataSource))
        else:
            return DataDriver(self.__dsources)  # Dsources themselves can be data sources or data drivers.

    def add_dsource(self, dsource):
        self.__dsources.append(dsource)

    def prepend_dsource(self, dsource):
        self._dsources.insert(0, dsource)

    @property
    def skipped(self):
        return self.__skipped

    @skipped.setter
    def skipped(self, flag):
        self.__skipped = flag

    @property
    def skip_code(self):
        return self.__skip_code

    def set_skip_code(self, code):
        self.__skipped = True
        self.__skip_code = code

    @property
    def picked(self):
        return not self.__unpicked

    def set_picked(self):
        self.__unpicked = False
        self.__unpick_code = None

    @property
    def unpicked(self):
        return self.__unpicked

    @unpicked.setter
    def unpicked(self, code):
        self.unpicked = True
        self.__unpick_code = code

    @property
    def unpick_code(self):
        return self.__unpick_code

    @property
    def skip_code(self):
        return self.__skip_code


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import traceback

from arjuna.unitee.types.root import Root
from arjuna.unitee.enums import *

from arjuna.unitee.reporter.result.types import FixtureResult
from arjuna.unitee.enums import ResultCodeEnum
from arjuna.unitee.exceptions import *

class Fixture(Root):
    def __init__(self, fixdef):
        super().__init__()
        self.name = fixdef.name
        self.qname = fixdef.qname
        self.type = fixdef.type
        self.func = fixdef.func
        self.__tvars = None
        self._result = None

    @property
    def tvars(self):
        return self.__tvars

    def __process_tvars_from_tobj(self, tvars):
        self.__tvars = tvars
        self.tvars.function['name'] = self.name
        self.tvars.function['qname'] = self.qname
        self.tvars.object['type'] = "FIXTURE"
        self.tvars.object['name'] = "Fixture"

    def execute(self, test_obj, tvars):
        from arjuna.unitee import Unitee
        self.__tvars = tvars
        result = None
        issue = False
        steps = None
        utvars = None
        try:
            self.unitee.state_mgr.get_current_thread_state().begin_recording(ExecutionContext.Fixture)
            utvars = self.__tvars.create_utvars()
            self.func(utvars)
        except StepResultEvent as sre:
            pass
            # We would assume the step exception added by Step itself
        except Exception as e:
            self.unitee.state_mgr.get_current_thread_state().add_step_exception(e, traceback.format_exc())

        self.tvars.context.update_from_context(utvars.context)
        steps = self.unitee.state_mgr.get_current_thread_state().end_recording()
        fresult = FixtureResult(test_obj, self)
        fresult.set_steps(steps)
        self._result = fresult
        return fresult

    def report(self):
        from arjuna.unitee import Unitee
        self.unitee.reporter.update_result(self._result)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc
import threading
import types

from arjuna.lib.reader.excel import *
from arjuna.lib.reader.ini import *
from arjuna.lib.reader.textfile import *
from arjuna.lib.thread import decorators
from arjuna.unitee.ddt.record import *
from arjuna.unitee.exceptions import *
from arjuna.lib.utils import sys_utils


class DataSource(metaclass=abc.ABCMeta):

    def __init__(self):
        self.lock = threading.RLock()
        self.name = None
        self.ended = False

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name

    @decorators.sync_method('lock')
    def terminate(self):
        self.ended = True

    def is_terminated(self):
        return self.ended

    def validate(self):
        pass

    @decorators.sync_method('lock')
    def next(self):
        if self.is_terminated():
            raise DataSourceFinished()

        try:
            data_record = self.get_next()
            if not data_record:
                raise StopIteration()
            record = self.process(data_record)
            if self.should_exclude(record):
                return self.next()
            else:
                return record
        except StopIteration as e:
            raise DataSourceFinished("Records Finished.")
        except Exception as e:
            raise DataSourceFinished(
                "Problem happened in getting next record. No further records would be provided.")

    def process(self, data_record):
        return data_record

    @abc.abstractmethod
    def get_next(self):
        pass

    @abc.abstractmethod
    def should_exclude(self, data_record):
        return False


class FileDataSource(DataSource):
    def __init__(self):
        super().__init__()
        self.reader = None

    def _set_reader(self, reader):
        self.reader = reader

    def get_next(self):
        return self.reader.next()

    def should_exclude(self, data_record):
        return False


class DsvFileDataSource(FileDataSource):
    def __init__(self, path, delimiter="\t"):
        super().__init__()
        self.reader = FileLine2ArrayReader(path, delimiter)
        self.headers = self.reader.get_headers()

    def process(self, data_record):
        return DataRecord(**dict(zip(self.headers, data_record)))


class IniFileDataSource(FileDataSource):
    def __init__(self, path):
        super().__init__()
        if path.lower().endswith("ini"):
            self.reader = IniFile2MapReader(path)
        else:
            raise Exception("Unsupported file extension for Ini reading.")

    def process(self, data_record):
        return DataRecord(**data_record)


class ExcelFileDataSource(FileDataSource):
    def __init__(self, path):
        super().__init__()
        if path.lower().endswith("xls"):
            self.reader = ExcelRow2ArrayReader(path)
            self.headers = self.reader.get_headers()
        else:
            raise Exception("Unsupported file extension for Excel reading.")

    def get_next(self):
        return self.reader.next()

    def should_exclude(self, data_record):
        return data_record.named_values().get('exclude', 'n') == 'y'

    def process(self, data_record):
        return DataRecord(**dict(zip(self.headers, data_record)))


class DummyDataSource(DataSource):
    MR = DataRecord()

    def __init__(self):
        super().__init__()
        self.done = False

    def get_next(self):
        if self.done:
            raise DataSourceFinished()
        else:
            self.done = True
            return DummyDataSource.MR

    def should_exclude(self, data_record):
        return False


class SingleDataRecordSource(DataSource):
    def __init__(self, record):
        super().__init__()
        self.record = record
        self.done = False

    def get_next(self):
        if not self.is_terminated() and not self.done:
            self.done = True
            return self.record
        else:
            raise DataSourceFinished()

    def should_exclude(self, data_record):
        return False


class DataArrayDataSource(DataSource):
    def __init__(self, records):
        super().__init__()
        self.records = records
        self.__iter = iter(self.records)

    def get_next(self):
        return next(self.__iter)

    def should_exclude(self, data_record):
        return False


class DataFunctionDataSource(DataSource):
    def __init__(self, func, *vargs, **kwargs):
        super().__init__()
        self.func = func
        self.vargs = vargs
        self.kwargs = kwargs
        try:
            self.__iter = iter(self.func(*self.vargs, **self.kwargs))
        except:
            raise Exception("data_function should return an object that is an iterable.")

    def get_next(self):
        obj = next(self.__iter)
        if isinstance(obj, DataRecord):
            return obj
        elif type(obj) is tuple or type(obj) is list:
            return DataRecord(*obj)
        elif type(obj) is dict:
            return DataRecord(**obj)
        else:
            return DataRecord(obj)

    def terminate(self):
        super().terminate()
        # self.ds.terminate()

    def should_exclude(self, data_record):
        return False


class DataClassDataSource(DataSource):
    def __init__(self, dclass, *vargs, **kwargs):
        super().__init__()
        self.klass = dclass
        self.vargs = vargs
        self.kwargs = kwargs
        try:
            self.__iter = iter(self.klass(*self.vargs, **self.kwargs))
        except:
            raise Exception("data_class should implement Python iteration protocol.")

    def get_next(self):
        obj = next(self.__iter)
        if isinstance(obj, DataRecord):
            return obj
        elif type(obj) is tuple or type(obj) is list:
            return DataRecord(*obj)
        elif type(obj) is dict:
            return DataRecord(**obj)
        else:
            return DataRecord(obj)

    def terminate(self):
        super().terminate()
        # self.ds.terminate()

    def should_exclude(self, data_record):
        return False


class MultiDataSource(DataSource):
    def __init__(self, dsource_defs):
        super().__init__()
        self.dsdefs = dsource_defs
        self.current_dsource = None
        self.def_iter = iter(self.dsdefs)

    def __init_next_source(self):
        try:
            dsdef = next(self.def_iter)
            self.current_dsource = dsdef.build()
        except:
            raise StopIteration()

    def get_next(self):
        if self.current_dsource is None:
            self.__init_next_source()

        try:
            return self.current_dsource.next()
        except StopIteration as e:
            self.__init_next_source()
            return self.get_next()

    def should_exclude(self, data_record):
        return False



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.unitee.state.loader import *
from arjuna.lib.utils import obj_utils

def init_module(id=None, *,
                priority=5,
                threads=1,
                evars=None,
                tags=None,
                bugs=None,
                drive_with=None,
                data_ref=None,
                exclude_if=None,

                name=None, author=None, idea=None, unstable=False,
                component=None, app_version=None,

                **kwargs):
    clocals = {i:j for i,j in locals().items()}
    from arjuna.tpi import Arjuna
    if obj_utils.callable(id):
        kallable = id
        del clocals['id']
        return Arjuna.get_unitee_instance().test_loader.register_tmodule(kallable, **clocals)
    else:
        def wrapper(kallable):
            return Arjuna.get_unitee_instance().test_loader.register_tmodule(kallable, **clocals)
        return wrapper

def test_function(id=None, *,
                  priority=5,
                  threads=1,
                  evars=None,
                  tags=None,
                  bugs=None,
                  drive_with=None,
                  data_ref=None,
                  exclude_if=None,

                  name=None, author=None, idea=None, unstable=False,
                  component=None, app_version=None,

                  **kwargs):
    clocals = {i:j for i,j in locals().items()}
    from arjuna.tpi import Arjuna
    if obj_utils.callable(id):
        kallable = id
        del clocals['id']
        return Arjuna.get_unitee_instance().test_loader.register_tfunc(kallable, **clocals)
    else:
        def wrapper(kallable):
            return Arjuna.get_unitee_instance().test_loader.register_tfunc(kallable, **clocals)
        return wrapper

def fixture(dec_name, kallable):
    from arjuna.tpi import Arjuna
    if obj_utils.callable(kallable):
        return Arjuna.get_unitee_instance().test_loader.register_fixture(dec_name, kallable)
    else:
        def wrapper(actual_kallable):
            msg = "You are decorating {} in {} module with @{} by providing one or more arguments."
            msg += "Remove the arguments and proceed."
            Arjuna.get_console().display_error(msg.format(
                    actual_kallable.__qualname__,
                    actual_kallable.__module__,
                    dec_name
                ))
            sys_utils.fexit()
        return wrapper


def skip_me(kallable):
    from arjuna.tpi import Arjuna
    if obj_utils.callable(kallable):
        return Arjuna.get_unitee_instance().test_loader.register_skip_func(kallable)
    else:
        def wrapper(actual_kallable):
            msg = "You are decorating {} in {} module with @skip by providing one or more arguments."
            msg += "Remove the arguments and proceed."
            Arjuna.get_console().display_error(msg.format(
                    actual_kallable.__qualname__,
                    actual_kallable.__module__
                ))
            sys_utils.fexit()
        return wrapper




import os
import sys
import time

def main(args=None):
    try:
        import signal
        import sys
        def signal_handler(sig, frame):
                print('Exiting...')
                sys.exit(0)
        signal.signal(signal.SIGINT, signal_handler)
        from arjuna import Arjuna
        Arjuna.launch(args and args or sys.argv)
    except Exception as e:
        # The following sleep is to accommodate a common IDE issue of
        # interspersing main exception with console output.
        time.sleep(0.5)
        msg = '''
    {0}
    Sorry. Looks like this is an error Arjuna couldn't handle.
    If Arjuna should handle this error, write to us: support@testmile.com
    {0}

    Message: {1}
        '''.format("-" * 70, str(e))
        print(msg)
        import traceback
        print(traceback.format_exc())



from enum import Enum, auto

from arjuna.setu.webclient.requester import SetuAgentRequester
from arjuna.lib.value import AnyRefValue


class ResponseCode(Enum):
    SUCCESS = auto()
    ERROR = auto()


class DefaultSetuResponse:

    def __init__(self, json_dict):
        self.__response_dict = json_dict

    def get_result(self):
        return ResponseCode[self.__response_dict["result"].upper()]

    def get_message(self):
        return self.__response_dict["emessage"]

    def get_trace(self):
        return self.__response_dict["etrace"]

    def get_data(self):
        return self.__response_dict["responseData"]

    def get_value_for_key(self, key):
        return AnyRefValue(self.get_data()[key])

    def get_value(self):
        return self.get_value_for_key("value")

    def get_value_for_value_attr(self):
        return self.get_value_for_key("value")

    def get_value_for_text(self):
        return self.get_value_for_key("text").as_string()

    def get_value_for_check_result(self):
        return self.get_value_for_key("checkResult")

    def get_value_for_element_setu_id(self):
        return self.get_value_for_key("elementSetuId").as_string()

    def get_value_for_testsession_setu_id(self):
        return self.get_value_for_key("testSessionSetuId").as_string()

    def get_value_for_config_setu_id(self):
        return self.get_value_for_key("configSetuId").as_string()

    def get_value_for_guiautomator_setu_id(self):
        return self.get_value_for_key("automatorSetuId").as_string()

    def get_gui_setu_id(self):
        return self.get_value_for_key("guiSetuId").as_string()

    def get_data_source_setu_id(self):
        return self.get_value_for_key("dataSourceSetuId").as_string()

    def add_data_item(self, name, value):
        if not self.get_data():
            self.__response_dict["responseData"] = {}
        self.__response_dict["responseData"][name] = value


class _DefaultSetuRequester:
    def __init__(self):
        self.__setu_url = "http://localhost:9000"
        self.__base_uri = "/setu"
        self.__rest_client = SetuAgentRequester(self.__setu_url)

    def post(self, request):
        return DefaultSetuResponse(self.__rest_client.post(self.__base_uri, request.as_json()))


class _DefaultSetuRequest:

    def __init__(self, component, setu_action_type):
        self.__request_body = dict()
        self.__request_body["component"] = component.name
        self.__request_body["action"] = setu_action_type.name
        self.__request_body["args"] = dict()

    def add_arg(self, name, value):
        self.__request_body["args"][name] = value

    def add_all_args(self, arg_dict):
        self.__request_body["args"].update(arg_dict)

    def as_json(self):
        return self.__request_body


class BaseSetuObject:
    SETU_CLIENT = _DefaultSetuRequester()

    def __init__(self):
        self.__setu_id = None
        self.__core_args = dict()

    def get_setu_id(self):
        return self.__setu_id

    def _set_setu_id(self, id):
        self.__setu_id = id

    def _set_test_session_setu_id_arg(self, id):
        self.__core_args["testSessionSetuId"] = id

    def _set_automator_setu_id_arg(self, id):
        self.__core_args["automatorSetuId"] = id

    def _set_gui_setu_id_arg(self, id):
        self.__core_args["guiSetuId"] = id

    def _set_self_setu_id_arg(self, id_name):
        self.__core_args[id_name] = self.get_setu_id()

    def _add_args(self, *vargs):
        for arg in vargs:
            self.__core_args[arg.get_name()] = arg.get_object()

    def __prepare_request_with_core_args(self, request):
        for name, value in self.__core_args.items():
            request.add_arg(name, value)

    def __prepare_request(self, request, *vargs):
        for arg in vargs:
            request.add_arg(arg.get_name(), arg.get_object())

    def _create_request(self, component, setu_action_type, *vargs):
        request = _DefaultSetuRequest(component, setu_action_type)
        self.__prepare_request_with_core_args(request)
        self.__prepare_request(request, *vargs)
        return request

    def _send_request(self, component, setu_action_type, *vargs):
        request = self._create_request(component, setu_action_type, *vargs)
        return self.SETU_CLIENT.post(request)


class SetuArg:

    def __init__(self, name, obj):
        self.__name = name
        self.__obj = obj

    def get_name(self):
        return self.__name

    def get_object(self):
        return self.__obj

    @staticmethod
    def arg(name, obj):
        return SetuArg(name, obj)

    @staticmethod
    def text_arg(obj):
        return SetuArg("text", obj)

    @staticmethod
    def value_arg(obj):
        return SetuArg("value", obj)

    @staticmethod
    def index_arg(obj):
        return SetuArg("index", obj)

    @staticmethod
    def config_arg(obj):
        return SetuArg("configSetuId", obj)

    @staticmethod
    def arg(name, obj):
        return SetuArg(name, obj)




from arjuna.lib.utils import sys_utils
from arjuna.lib.utils import etree_utils
from arjuna.unitee.types.containers import *
from arjuna.unitee.test.defs.fixture import *

ALLOWED_CONFIG_ATTRS = {'name', 'value'}
MANDATORY_CONFIG_ATTRS = {'name', 'value'}
ALLOWED_FIXTURE_ATTRS = {'type', 'module', 'func'}

ALLOWED_STAGE_ATTRS = {'name', 'threads'}

ALLOWED_GROUP_ATTRS = {'name', 'threads'}
MANDATORY_GROUP_ATTRS = {'name'}

ALLOWED_GCONF_ATTRS = {'name'}
MANDATORY_GCONF_ATTRS = ALLOWED_GCONF_ATTRS

ALLOWED_PATTERN_ATTRS = {'type', 'pattern'}
MANDATORY_PATTERN_ATTRS = ALLOWED_PATTERN_ATTRS

ALLOWED_RULE_ATTRS = {'if', 'in'}
MANDATORY_RULE_ATTRS = {'if'}

VALID_FIXTURES = {
    "session" : {"init_session", "end_session", "init_each_stage", "end_each_stage"},
    "stage" : {"init_stage", "end_stage", "init_each_group", "end_each_group"},
    "group" : {"init_group", "end_group", "init_each_module", "end_each_module"}
}

converter_map = {
    'int': int,
    'str': str,
    'float': float,
    'bool': bool
}

def display_err_and_exit(conf_type, conf_path, msg):
    from arjuna import ArjunaCore
    console = ArjunaCore.console
    console.display_error(msg + " Fix {} template file: {}".format(conf_type, conf_path))
    sys_utils.fexit()


def validate_config_xml_child(conf_type, session_path, config_node):
    if config_node.tag.lower() not in {'arjuna_option', 'user_option'}:
        display_err_and_exit(conf_type, session_path, ">>config<< element can only contain one or more children of type >>arjuna_option<< or >>user_option<<.")
    for k,v in config_node.attrib.items():
        if k.lower() not in ALLOWED_CONFIG_ATTRS:
            display_err_and_exit(conf_type, session_path, ">>{}<< element attributes can only be set as: {}.".format(config_node.tag, ALLOWED_CONFIG_ATTRS))
    if not MANDATORY_CONFIG_ATTRS.issubset(etree_utils.convert_attribs_to_cidict(config_node)):
        display_err_and_exit(conf_type, session_path, ">>{}<< element must define >>name<< and >>value<< attributes.")


def add_config_node_to_configuration(conf_type, config_container, config_node):
    config_attrs = CIStringDict(config_node.attrib)
    if config_node.tag.lower() == "arjuna_option":
        config_container.set_arjuna_option(config_attrs['name'], config_attrs['value'])
    else:
        config_container.set_user_option(config_attrs['name'], config_attrs['value'])


def validate_fixture_xml_child(conf_type, object_type, session_path, fixture_node):
    if fixture_node.tag.lower() != 'fixture':
        display_err_and_exit(conf_type, ">>fixtures<< element can only contain one or more children of type >>fixture<<.")
    for k,v in fixture_node.attrib.items():
        if k.lower() not in ALLOWED_FIXTURE_ATTRS:
            display_err_and_exit(conf_type, ">>fixtures<< attributes can only be set as: {}.".format(ALLOWED_FIXTURE_ATTRS))
    if not ALLOWED_FIXTURE_ATTRS.issubset(fixture_node.attrib):
        display_err_and_exit(conf_type, ">>fixtures<< element must define >>type<<, >>module<< and >>func<< attributes.")
    for k,v in fixture_node.attrib.items():
        if fixture_node.attrib['type'].lower() not in VALID_FIXTURES[object_type.lower()]:
            display_err_and_exit(
                conf_type,
                session_path,
                ">>fixtures<< element inside >>{}<< can only only contain one of these fixture types: {}".format(
                            object_type,
                            VALID_FIXTURES[object_type.lower()]
                )
            )

def add_fixture_node_to_fixdefs(fixture_defs, fixture_node):
    fixture_attrs = CIStringDict(fixture_node.attrib)

    ConfiguredFixtureHelper.configure_fixture(
        fixture_defs,
        fixture_attrs['type'].upper(),
        fixture_attrs['module'].strip(),
        fixture_attrs['func'].strip()
    )


def validate_stage_xml_child(conf_type, session_path, stage_node):
    if stage_node.tag.lower() != 'stage':
        display_err_and_exit(conf_type, session_path, ">>stages<< element can only contain one or more children of type >>stage<<.")
    for k,v in stage_node.attrib.items():
        if k.lower() not in ALLOWED_STAGE_ATTRS:
            display_err_and_exit(conf_type, session_path, ">>stage<< attributes can only be set as: {}.".format(ALLOWED_STAGE_ATTRS))


def validate_group_xml_child(conf_type, session_path, group_node):
    if group_node.tag.lower() != 'group':
        display_err_and_exit(conf_type, session_path, ">>groups<< element can only contain one or more children of type >>group<<.")
    for k,v in group_node.attrib.items():
        if k.lower() not in ALLOWED_GROUP_ATTRS:
            display_err_and_exit(conf_type, session_path, ">>group<< attributes can only be set as: {}.".format(ALLOWED_GROUP_ATTRS))
    if not MANDATORY_GROUP_ATTRS.issubset(etree_utils.convert_attribs_to_cidict(group_node)):
        display_err_and_exit(conf_type, session_path, ">>group<< element must define these attributes: {}".format(MANDATORY_GROUP_ATTRS))


def validate_gconf_xml_child(conf_type, groups_path, gconf_node):
    if gconf_node.tag.lower() != 'group':
        display_err_and_exit(conf_type, groups_path, ">>groups<< element can only contain one or more children of type >>group<<.")
    for k,v in gconf_node.attrib.items():
        if k.lower() not in ALLOWED_GCONF_ATTRS:
            display_err_and_exit(conf_type, groups_path, ">>group<< attributes can only be set as: {}.".format(ALLOWED_GCONF_ATTRS))
    if not MANDATORY_GCONF_ATTRS.issubset(etree_utils.convert_attribs_to_cidict(gconf_node)):
        display_err_and_exit(conf_type, groups_path, ">>group<< element must define these attributes: {}".format(MANDATORY_GCONF_ATTRS))

def validate_pattern_xml_child(conf_type, group_name, groups_path, pattern_node):
    for k, v in pattern_node.attrib.items():
        if k.lower() not in ALLOWED_PATTERN_ATTRS:
            display_err_and_exit(conf_type, groups_path,
                                 ">>patten<< attributes can only be set as: {}. Check group: {}".format(ALLOWED_PATTERN_ATTRS, group_name))
    if not MANDATORY_PATTERN_ATTRS.issubset(etree_utils.convert_attribs_to_cidict(pattern_node)):
        display_err_and_exit(conf_type, groups_path,
                             ">>pattern<< element must define these attributes: {}. Check group: {}".format(MANDATORY_PATTERN_ATTRS, group_name))

def validate_rule_xml_child(conf_type, group_name, groups_path, rule_node):
    for k, v in rule_node.attrib.items():
        if k.lower() not in ALLOWED_RULE_ATTRS:
            display_err_and_exit(conf_type, groups_path,
                                 ">>patten<< attributes can only be set as: {}. Check group: {}".format(ALLOWED_RULE_ATTRS, group_name))
    if not MANDATORY_RULE_ATTRS.issubset(etree_utils.convert_attribs_to_cidict(rule_node)):
        display_err_and_exit(conf_type, groups_path,
                             ">>pattern<< element must define these attributes: {}. Check group: {}".format(MANDATORY_RULE_ATTRS, group_name))


from arjuna.tpi.enums import ArjunaOption, SetuActorMode
from arjuna.setu.dispatcher.setu_actor_requester import SetuActorRequester
from arjuna.tpi.enums import GuiAutomatorName
from .element import GuiElementDispatcher
from functools import partial

class GuiAutomatorDispatcher:

    def __init__(self, config, setu_id):
        self.__config = config
        self.__setu_id = setu_id
        self.__dispatcher = None
        self.__automator_name = config.setu_config.get_gui_automator_name()
        self.__actor_mode = config.setu_config.get_guiauto_actor_mode()

        if self.__actor_mode == SetuActorMode.REMOTE:
            actor_url = config.setu_config.value(ArjunaOption.SETU_GUIAUTO_ACTOR_URL)
            if (self.__automator_name == GuiAutomatorName.SELENIUM):
                from arjuna.setuext.guiauto.dispatcher.selenium.remote.driver import SeleniumDriver
                self.__dispatcher = SeleniumDriver(setu_id, SetuActorRequester(actor_url))
            else:
                raise Exception("{} automator is not aupported.".format(self.__automator_name.name))
        else:
            if (self.__automator_name == GuiAutomatorName.SELENIUM):
                from arjuna.setuext.guiauto.dispatcher.selenium.local.driver import SeleniumDriver
                self.__dispatcher = SeleniumDriver(setu_id)
            else:
                raise Exception("{} automator is not aupported.".format(self.__automator_name.name))

    def create_gui_element_dispatacher(self, element_setu_id):
        return GuiElementDispatcher(
            self.__config,
            self.__automator_name,
            self.__dispatcher, 
            element_setu_id
        )

    def __getattr__(self, attr):
        return partial(vars(self.__dispatcher.__class__)[attr], self.__dispatcher)

    # Write __whatever__ override and check method delegation valid or not.


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.tpi import Arjuna
from arjuna.lib.reader.hocon import *
from arjuna.unitee.types.containers import *
from arjuna.unitee.types.root import *

from arjuna.lib.utils import sys_utils
from arjuna.lib.utils import etree_utils
from arjuna.unitee.utils import run_conf_utils


class GroupDef(Root):

    def __init__(self, sdef, stage_def, id, group_xml_from_session):
        super().__init__()
        self.unitee = Arjuna.get_unitee_instance()
        self.sdef = sdef
        self.stage_def = stage_def
        self.id = id
        self.root = group_xml_from_session
        self.__gconf = None
        self.__iter = None
        self.__process()

        self.__mnames = []
        self.__mod_fname_map = {}

    @property
    def fixture_defs(self):
        return self.__gconf.fixture_defs

    @property
    def rules(self):
        return self.__gconf.rules

    @property
    def threads(self):
        return self.__gconf.threads
        
    def __getattr__(self, item):
        return vars(self.__gconf)[item]

    def __process(self):
        def display_err_and_exit(msg):
            self.console.display_error((msg + " Fix session template file: {}").format(self.sdef.fpath))
            sys_utils.fexit()

        group_attrs = etree_utils.convert_attribs_to_cidict(self.root)
        self.name = group_attrs['name'].strip()
        if not self.name:
            display_err_and_exit(
                ">>name<< attribute in group definition can not be empty.")

        if self.name not in self.unitee.groups:
            display_err_and_exit(
                ">>name<< attribute in group definition is pointing to a non-existing group '{}'.".format(self.name))
        else:
            self.__gconf = self.unitee.groups[self.name]

        threads_err_msg = ">>threads<< attribute in stage definition can be integer >=1."
        if "threads" in group_attrs:
            self.threads = group_attrs['threads'].strip()
            try:
                self.threads = int(self.threads)
            except:
                display_err_and_exit(threads_err_msg)
            else:
                if self.threads <=0:
                    display_err_and_exit(threads_err_msg)

        node_dict = etree_utils.convert_to_cidict(self.root)

        for child_name, child in node_dict.items():
            if child.tag == 'evars':
                evars = child
                for child in evars:
                    run_conf_utils.validate_config_xml_child("session", self.sdef.fpath, child)
                    run_conf_utils.add_config_node_to_configuration("session", self.__gconf.evars, child)
            else:
                display_err_and_exit("Unexpected element >>{}<< found in >>group<< definition in session file.".format(child.tag))

    def pick(self):
        mnames = self.unitee.testdb.get_mnames()

        # Pick what you can
        group_tfunc_count = 0
        for mname in mnames:
            if self.picker.include_module(mname):
                mdef = self.unitee.testdb.get_mdef(mname)
                mdef.set_picked()
                scheduled_fnames = []
                tfunc_count = 0
                for fname in mdef.get_func_names():
                    if self.picker.include_func(fname):
                        fdef = mdef.get_fdef(fname)
                        scheduled_fnames.append(fname)
                        tfunc_count += 1
                        fdef.set_picked()
                    else:
                        continue
                if tfunc_count != 0:
                    self.__mnames.append(mname)
                    self.__mod_fname_map[mname] = scheduled_fnames
                    group_tfunc_count += tfunc_count
            else:
                continue

        return group_tfunc_count

    def get_picked_mnames(self):
        return [mname for mname in self.__mnames if self.unitee.testdb.should_run_module(mname)]

    def get_schedule_module_map(self):
        out_map = {}
        for mqname, fqnames in self.__mod_fname_map.items():
            if self.unitee.testdb.should_run_module(mqname):
                out_map[mqname]= []
                mdef = self.unitee.testdb.get_mdef(mqname);
                for fqname in fqnames:
                    if mdef.should_run_func(fqname):
                        out_map[mqname].append(fqname)
        return out_map


from enum import Enum, auto
from arjuna.client.core.action import *
from arjuna.client.core.config import ArjunaComponent
from arjuna.client.core.connector import BaseSetuObject, SetuArg
from .component import GuiAutoComponentFactory


class GuiAutomatorName(Enum):
    SELENIUM = auto()
    APPIUM = auto()


class AbstractAppAutomator(BaseSetuObject):

    def __init__(self, config=None):
        super().__init__()
        self.__dom_root = None
        self.__main_window = None
        self.__browser = None
        self.__auto_context = None
        self.__test_session = None
        self.__config = None
        if config:
            self._set_config(config)
            self.__test_session = config.get_test_session()
            self._set_test_session_setu_id_arg(self.__test_session.get_setu_id())
            self.__config = config

    def _set_automation_context(self, context):
        self.__auto_context = context

    def get_automation_context(self):
        return self.__auto_context

    def is_gui(self):
        return False

    def __take_element_finder_action(self, setu_action_type, *setu_args):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, setu_action_type, *setu_args)
        return response.get_value_for_element_setu_id()

    def __create_generic_element(self, setu_action_type, *with_locators):
        if with_locators:
            locators = [l.as_map() for l in with_locators]
            return self.__take_element_finder_action(setu_action_type, SetuArg.arg("locators", locators))
        else:
            return self.__take_element_finder_action(setu_action_type)

    def Element(self, *with_locators):
        elem_setu_id = self.__create_generic_element(GuiAutoActionType.CREATE_ELEMENT, *with_locators)
        return GuiAutoComponentFactory.Element(self.__test_session, self, elem_setu_id)

    def MultiElement(self, *with_locators):
        elem_setu_id = self.__create_generic_element(GuiAutoActionType.CREATE_MULTIELEMENT, *with_locators)
        return GuiAutoComponentFactory.MultiElement(self.__test_session, self, elem_setu_id)

    def DropDown(self, *with_locators):
        elem_setu_id = self.__create_generic_element(GuiAutoActionType.CREATE_DROPDOWN, *with_locators)
        return GuiAutoComponentFactory.DropDown(self.__test_session, self, elem_setu_id)

    def RadioGroup(self, *with_locators):
        elem_setu_id = self.__create_generic_element(GuiAutoActionType.CREATE_RADIOGROUP, *with_locators)
        return GuiAutoComponentFactory.RadioGroup(self.__test_session, self, elem_setu_id)

    def Alert(self):
        elem_setu_id = self.__create_generic_element(GuiAutoActionType.CREATE_ALERT)
        return GuiAutoComponentFactory.Alert(self.__test_session, self, elem_setu_id)

    def Frame(self, *with_locators):
        return self.DomRoot().Frame(*with_locators)

    def ChildWindow(self, *with_locators):
        return self.MainWindow().ChildWindow(*with_locators)

    def LatestChildWindow(self):
        return self.MainWindow().LatestChildWindow()

    def close_all_child_windows(self):
        self.MainWindow().close_all_child_windows()

    def MainWindow(self):
        return self.__main_window

    def _set_main_window(self, window):
        self.__main_window = window

    def get_config(self):
        return self.__config

    def _set_config(self, config):
        self.__config = config

    def DomRoot(self):
        return self.__dom_root

    def _set_dom_root(self, root):
        self.__dom_root = root

    def get_test_session(self):
        return self.__test_session

    def Browser(self):
        return self.__browser

    def _set_browser(self, browser):
        self.__browser = browser

    def enable_slow_motion(self, on, interval=None):
        args = [SetuArg.arg("on", on)]
        if interval:
            args.append(SetuArg.arg("interval", interval))
        self._send_request(GuiAutoActionType.SET_SLOMO, *args)

    def execute_javascript(self, script):
        self._send_request(GuiAutoActionType.BROWSER_EXECUTE_JAVASCRIPT, SetuArg.arg("script", script))


class DefaultGuiAutomator(AbstractAppAutomator):

    def __init__(self, config, extendedConfig=None):
        super().__init__(config)
        self.__extended_config = extendedConfig
        self.__launch()

    def __launch(self):
        response = None
        if self.__extended_config:
            response = self._send_request(
                ArjunaComponent.GUI_AUTOMATOR,
                GuiAutoActionType.LAUNCH_AUTOMATOR,
                SetuArg.config_arg(self.get_config().get_setu_id()),
                SetuArg.arg("extendedConfig", self.__extended_config)
            )
        else:
            response = self._send_request(
                ArjunaComponent.GUI_AUTOMATOR,
                GuiAutoActionType.LAUNCH_AUTOMATOR,
                SetuArg.config_arg(self.get_config().get_setu_id())
            )

        self._set_setu_id(response.get_value_for_guiautomator_setu_id())
        self._set_self_setu_id_arg("automatorSetuId")

        win_response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DEFINE_MAIN_WINDOW)
        self._set_main_window(GuiAutoComponentFactory.MainWindow(self.get_test_session(), self, win_response.get_value_for_element_setu_id()))

        self._set_dom_root(GuiAutoComponentFactory.DomRoot(self.get_test_session(), self))
        self._set_browser(GuiAutoComponentFactory.Browser(self.get_test_session(), self))

    def quit(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.QUIT_AUTOMATOR)



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc
from enum import Enum, auto
from arjuna.lib.utils import obj_utils

class TestObjectsEnum:
    MODULES = auto()
    FUNCTIONS = auto()

class TestObjects:

    def __init__(self, objects, ctype):
        self.__objects = objects
        self.__ctype = ctype

    @property
    def type(self):
        return self.__ctype

    @property
    def objects(self):
        return self.__objects

    def __iter__(self):
        return iter(self.objects)

class modules(TestObjects):

    def __init__(self, module, *modules):
        super().__init__([module] + list(modules), TestObjectsEnum.MODULES)

        for module in self.objects:
            if not obj_utils.is_module(module):
                raise Exception("modules() call expects all arguments to be module objects. {} is not a module".format(module))

class functions(TestObjects):

    def __init__(self, func, *funcs):
        super().__init__([func] + list(funcs), TestObjectsEnum.FUNCTIONS)

        for func in self.objects:
            if not obj_utils.is_function(func):
                raise Exception("functions() call expects all arguments to be function (unbound, non-inner) objects. {} is not a function".format(func))

class DepUnit(metaclass=abc.ABCMeta):

    def __init__(self, name):
        self.__ignore = False
        self.__name = name

    @property
    def name(self):
        return self.__name

    def should_ignore(self):
        return self.__ignore

    def set_ignored(self):
        self.__ignore = True

    @abc.abstractmethod
    def evaluate(self, state):
        pass

class ProblemDepUnit(DepUnit):
    def __init__(self, name):
        super().__init__(name)

    def evaluate(self, state):
        if not self.should_ignore():
            state.raise_on_problem(self.name)

class ProlemModule(ProblemDepUnit):
    def __init__(self, mname):
        super().__init__(mname)

class ProblemFunction(ProblemDepUnit):
    def __init__(self, mname, name):
        super().__init__(name)
        self.mname = mname

class Dependency:

    def __init__(self, target):
        self._target = target
        self._depunits = []

    def is_mdep(self):
        return self._target.type == TestObjectsEnum.MODULES

    def is_fdep(self):
        return self._target.type == TestObjectsEnum.FUNCTIONS

    def append(self, du):
        self._depunits.append(du)

    def __iter__(self):
        return iter(self._depunits)

    def evaluate(self, state):
        for du in self._depunits:
            du.evaluate(state)

class problem_in(Dependency):
    def __init__(self, callables_obj):
        super().__init__(callables_obj)
        if not isinstance(callables_obj, TestObjects):
            raise Exception(
                "problem_in() call expects its argument to be a TestObjects object: either modules() or functions(). Provided: {}".format(
                    callables_obj))
        self._callables = callables_obj

    def process(self):
        klass = None
        if self.is_mdep():
            for obj in self._target:
                self.append(ProlemModule(obj.__name__))
        else:
            for obj in self._target:
                self.append(ProblemFunction(obj.__module__, obj.__name__))



from arjuna.setuext.guiauto.impl.base.conditions import DynamicCaller, CommandCondition

class GuiElementContainerConditions:

    def __init__(self, container):
        self.__container = container

    def PresenceOfElement(self, gui_element):
        caller = DynamicCaller(
            self.__container._find,  
            self.__container.dispatcher.find_element,
            gui_element
        )
        return CommandCondition(caller)   

    def PresenceOfMultiElement(self, gui_element):
        caller = DynamicCaller(
            self.__container._find, 
            self.__container.dispatcher.find_multielement,
            gui_element
        )
        return CommandCondition(caller)  


import abc

from arjuna.setu.types import SetuConfiguredObject

from .container_conditions import GuiElementContainerConditions


class ElementContainer(SetuConfiguredObject, metaclass=abc.ABCMeta):
    def __init__(self, config):
        super().__init__(config)
        self.element_map = {}
        self.melement_map = {}
        self.__container_conditions = GuiElementContainerConditions(self)

    def _add_element(self, setu_id, element):
        self.element_map[setu_id] = element

    def _add_multielement(self, setu_id, melement):
        self.melement_map[setu_id] = melement

    def get_element_for_setu_id(self,id):
        return self.element_map[id]

    def get_multielement_for_setu_id(self,id):
        return self.melement_map[id]

    @abc.abstractmethod
    def _create_element_flat_or_nested(self, locator_meta_data):
        pass

    @abc.abstractmethod
    def _create_multielement_flat_or_nested(self, locator_meta_data):
        pass

    def create_element(self, locator_meta_data):
        elem = self._create_element_flat_or_nested(locator_meta_data)
        elem.dispatcher_creator = self.dispatcher_creator
        self._add_element(elem.get_setu_id(), elem)
        return elem

    def create_multielement(self, locator_meta_data):
        element = self._create_multielement_flat_or_nested(locator_meta_data)
        element.dispatcher_creator = self.dispatcher_creator
        self._add_multielement(element.get_setu_id(), element)
        return element

    def create_dropdown(self, locator_meta_data):
        from arjuna.setuext.guiauto.impl.element.dropdown import GuiWebSelect
        select = GuiWebSelect(self, locator_meta_data)
        self._add_element(select.get_setu_id(), select)
        return select

    def create_radiogroup(self, locator_meta_data):
        from arjuna.setuext.guiauto.impl.element.radio_group import GuiWebRadioGroup
        rg = GuiWebRadioGroup(self, locator_meta_data)
        rg.dispatcher_creator = self.dispatcher_creator
        self._add_element(rg.get_setu_id(), rg)
        return rg

    def _find(self, dispatcher_call, gui_element):
        found = False
        for locator in gui_element.get_locator_meta_data().locators: 
            try:
                instance_count = dispatcher_call(gui_element.get_setu_id(), locator.ltype.name, locator.lvalue)
                return locator.ltype.name, locator.lvalue, instance_count
            except Exception as e:
                print(e)
                continue
        if not found:
            raise Exception("Could not locate elements with locator(s): {}".format(gui_element.get_locator_meta_data().locators))

    def wait_until_element_found(self, gui_element):
        return self.__container_conditions.PresenceOfElement(gui_element).wait()

    def wait_until_multielement_found(self, gui_element):
        return self.__container_conditions.PresenceOfMultiElement(gui_element).wait()

    def find_multielement(self, gui_element):
        locator_type, locator_value, instance_count = self.wait_until_multielement_found(gui_element)
        gui_element.set_found_with(locator_type, locator_value)
        gui_element.set_instance_count(instance_count)

    def find_element(self, gui_element):
        locator_type, locator_value, __ = self.wait_until_element_found(gui_element)
        gui_element.set_found_with(locator_type, locator_value)



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .base import *

from arjuna.unitee.types.containers import *
from arjuna.unitee.test.defs.deps import *
from arjuna.unitee.ddt.source import *

class FuncDef(TestSourceDef):
    def __init__(self, mdef, t):
        super().__init__()
        self.mdef = mdef
        self.pkg = t.pkg
        self.module = t.module
        self.name = t.name
        self.qname = t.qname

        self.func = t.obj
        self.__threads = 1

        self.tvars = TestVars()
        self.tvars.info.function = FunctionInfo()
        self.tvars.info.function.meta['name'] = self.name
        self.tvars.info.function.meta['qname'] = self.qname

        self.__dependency = None
        self.__data_driver = None

    @property
    def threads(self):
        return self.__threads

    @threads.setter
    def threads(self, count):
        self.__threads = count

    def set_evars(self, variables):
        if variables:
            self.tvars.evars.update(variables.get_dict())

    @property
    def data_driver(self):
        return self.__data_driver

    @data_driver.setter
    def data_driver(self,data_driver):
        self.__data_driver = data_driver

    @property
    def data_source(self):
        if self.__data_driver:
            return self.__data_driver.build()
        else:
            return DummyDataSource()

    @property
    def dependency(self):
        return self.__dependency

    @dependency.setter
    def dependency(self, dep):
        if not dep: return
        if not isinstance(dep, Dependency):
            raise Exception ("exclude_if attribute expects the value to be a dependency object.")
        elif not dep.is_fdep():
            raise Exception("exclude_if attribute in @test_function expects the dependency object of functions().")
        else:
            self.__dependency = dep

    def enumerate(self):
        self.console.marker(50, 'f')
        self.console.display("Function:", self.name)
        self.console.display(self.tvars)
        self.console.marker(50, 'f')




'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc
import datetime

from arjuna.tpi import Arjuna
from arjuna.unitee import Unitee
from arjuna.unitee.enums import *
from arjuna.unitee.reporter.result.types import *
from arjuna.unitee.exceptions import *
from arjuna.unitee.selection.rules.common.exceptions import *

class TestObject(metaclass=abc.ABCMeta):
    def __init__(self, totype):
        self.unitee = Arjuna.get_unitee_instance()
        self.__type = totype
        self.__tvars = None
        self.__thcount = 1
        self.__thname = None
        self.__iter = None
        self.__before_fixtures = []
        self.__after_fixtures = []
        self.__children = None
        self.__state = None
        self.__reported = False

    @property
    def reported(self):
        return self.__reported

    @reported.setter
    def reported(self, flag):
        self.__reported = flag

    @property
    def state(self):
        return self.__state

    @state.setter
    def state(self, state):
        self.__state = state

    @property
    def type(self):
        return self.__type

    @property
    def tvars(self):
        return self.__tvars

    @tvars.setter
    def tvars(self, tvars):
        self.__tvars = tvars

    @property
    def thcount(self):
        return self.__thcount

    @thcount.setter
    def thcount(self, count):
        self.__thcount = count

    @property
    def thname(self):
        return self.__thname

    @thname.setter
    def thname(self, name):
        self.__thname = name

    @property
    def children(self):
        return self.__children

    @children.setter
    def children(self, cl):
        self.__children = cl
        self.__iter = iter(self.__children)

    def next(self):
        try:
            child = next(self.__iter)
            child.update_dynamic_tvars(self.tvars)
            return child
        except StopIteration:
            raise SubTestsFinished()

    def update_dynamic_tvars(self, tvars):
        self.tvars.evars.update(tvars.evars)

    @property
    def before_fixtures(self):
        return self.__before_fixtures

    def append_before_fixture(self, fix):
        if fix:
            self.__before_fixtures.append(fix)

    @property
    def after_fixtures(self):
        return self.__after_fixtures

    def append_after_fixture(self, fix):
        if fix:
            self.__after_fixtures.append(fix)

    def run(self, base_tvars=None):
        if base_tvars:
            self.tvars.context = base_tvars.context.clone()
            self.tvars.evars.update(base_tvars.evars)
            self.tvars.runtime.update(base_tvars.runtime)

        bstamp = datetime.datetime.now().timestamp()
        if self.type not in {TestObjectTypeEnum.GSlot, TestObjectTypeEnum.MSlot}:
            info = Info(InfoType.STARTED, self)
            info.btstamp = bstamp
            self.unitee.reporter.update_info(info)

        try:
            self.should_i_surrender()
        except DependencyNotMet as e:
            self.exclude(ResultCodeEnum.DEPENDENCY_NOT_MET, e.iid, ResultCodeEnum.PARENT_DEPENDENCY_NOT_MET)
            self.report_finish_info(bstamp)
            return
        except RuleNotMet as e:
            self.surrender(ResultCodeEnum.SURRENDERED_RULES_NOT_MET, ResultCodeEnum.PARENT_SURRENDERED)
            self.report_finish_info(bstamp)
            return

        # Some test objects have multiple before fixtures e.f. init_each_group and init_group
        for before_fixture in self.before_fixtures:
            fresult = before_fixture.execute(self, self.tvars.clone())
            before_fixture.report()
            self.tvars.context = before_fixture.tvars.context.clone()
            #self.tvars.evars.update(before_fixture.tvars.evars)
            self.tvars.runtime.update(before_fixture.tvars.runtime)
            if fresult.has_issues():
                self.exclude(ResultCodeEnum["{}_{}".format(fresult.rtype, before_fixture.type.name)], fresult.iid)
                self.report_finish_info(bstamp)
                return

        self._execute()

        if self.type != TestObjectTypeEnum.Test:
            if self.type not in {TestObjectTypeEnum.GSlot, TestObjectTypeEnum.MSlot}:
                for after_fixture in self.after_fixtures:
                    fresult = after_fixture.execute(self, self.tvars.clone())
                    if fresult.has_issues():
                        if self.type not in {TestObjectTypeEnum.GSlot, TestObjectTypeEnum.MSlot} and not self.reported:
                            result = TestObjectResult(
                                self, rtype=fresult.rtype,
                                rcode=ResultCodeEnum["{}_{}".format(fresult.rtype, after_fixture.type.name)],
                                iid=fresult.iid)
                            self.report(result)
                    else:
                        if self.type not in {TestObjectTypeEnum.GSlot, TestObjectTypeEnum.MSlot} and not self.reported:
                            result = TestObjectResult(self, rtype=ResultTypeEnum.PASS,
                                                      rcode=ResultCodeEnum.PASS_ALL_CHILD_TEST_OBJECTS)
                            self.report(result)
                    after_fixture.report()

                if not self.reported:
                    if self.state.has_failures():
                        result = TestObjectResult(self, rtype=ResultTypeEnum.FAIL,
                                                  rcode=ResultCodeEnum.FAIL_CHILD_TEST_OBJECT)
                        self.report(result)
                    elif self.state.has_errors():
                        result = TestObjectResult(self, rtype=ResultTypeEnum.ERROR,
                                                  rcode=ResultCodeEnum.ERROR_CHILD_TEST_OBJECT)
                        self.report(result)
                    elif self.state.has_exclusions():
                        result = TestObjectResult(self, rtype=ResultTypeEnum.EXCLUDED_CHILDREN,
                                                  rcode=ResultCodeEnum.EXCLUDED_CHILD_TEST_OBJECT)
                        self.report(result)

        if self.type not in {TestObjectTypeEnum.Test, TestObjectTypeEnum.GSlot, TestObjectTypeEnum.MSlot} and not self.reported:
            result = TestObjectResult(self, rtype=ResultTypeEnum.PASS,
                                      rcode=ResultCodeEnum.PASS_ALL_CHILD_TEST_OBJECTS)
            self.report(result)

        self.report_finish_info(bstamp)

    def report_finish_info(self, bstamp):
        if self.type not in {TestObjectTypeEnum.GSlot, TestObjectTypeEnum.MSlot}:
            info = Info(InfoType.FINISHED, self)
            info.btstamp = bstamp
            info.etstamp = datetime.datetime.now().timestamp()
            info.exec_time = info.etstamp - info.btstamp
            self.unitee.reporter.update_info(info)

    @abc.abstractmethod
    def _execute(self):
        pass

    @abc.abstractmethod
    def _populate_tvars(self, base_tvars):
        pass

    @abc.abstractmethod
    def load(self, ptvars):
        pass

    def report(self, result):
        self.unitee.reporter.update_result(result)
        self.reported = True

    def report_excluded(self, exclusion_type, issue_id):
        t = TestObjectResult(self, rtype=ResultTypeEnum.EXCLUDED, rcode=exclusion_type, iid=issue_id)
        self.report(t)

    def report_surrendered(self, surrender_type):
        t = TestObjectResult(self, rtype=ResultTypeEnum.SURRENDERED, rcode=surrender_type)
        self.report(t)

    def exclude(self, exclusion_type, issue_id, child_exclusion_type=None):
        if not child_exclusion_type:
            child_exclusion_type = exclusion_type
        if self.children:
            for child in self.children:
                child.exclude(child_exclusion_type, issue_id)
        self.report_excluded(exclusion_type, issue_id)

    def surrender(self, surrender_type, child_surrender_type=None):
        if not child_surrender_type:
            child_surrender_type = surrender_type
        if self.children:
            for child in self.children:
                child.surrender(child_surrender_type)
        self.report_surrendered(surrender_type)

    def _evaluate_dependency(self):
        pass

    def _evaluate_rules(self):
        pass

    def should_i_surrender(self):
        self._evaluate_dependency()
        self._evaluate_rules()



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os
import shutil
import re

def get_canonical_path(file_path):
    return os.path.abspath(file_path)


def create_file_path(parent_path, child_parts, file_name, extension):
    dir_path = os.pathsep.join([
        parent_path,
        os.pathsep.join(child_parts)
    ])

    os.makedirs(dir_path)
    return os.pathsep.join(dir_path, file_name + extension)


def is_file(path, d=None):
    p = path
    if d:
        p = os.path.join(d, p)
    return os.path.exists(p) and os.path.isfile(p)


def is_dir(path):
    return os.path.exists(path) and os.path.isdir(path)


def is_absolute_path(path):
    return os.path.isabs(path)

def validate_file(path):
    if not os.path.exists(path):
        raise Exception("File does not exist.")
    elif os.path.isfile(path):
        raise Exception("Not a file.")


def validate_dir(path):
    if not os.path.exists(path):
        raise Exception("Directory does not exist.")
    elif os.path.isfile(path):
        raise Exception("Not a directory.")


def get_file_modified_time_stamp(path):
    return os.path.getmtime(path)


def get_latest_file_pathfrom_dir(path):
    check = 0
    last = None
    for f in os.listdir(path):
        m = get_file_modified_time_stamp(os.pathsep.join(path, f))
        if m > check:
            check = m
            last = m

    return last


def delete_dir(dir_path):
    validate_dir(dir_path)
    shutil.rmtree(dir_path)

def delete_dir_if_exists(path):
    if not os.path.exists(path): return
    if not is_dir(path):
        raise Exception("{} is not a directory.".format(path))
    shutil.rmtree(path)

def copy_file(src_file, dest_file):
    validate_file(src_file)
    validate_dir(os.path.dirname(dest_file))
    shutil.copy2(src_file, dest_file)


def copy_file_to_dir(src_file, dest_dir):
    validate_file(src_file)
    validate_dir(dest_dir)
    shutil.copy2(src_file, dest_dir)


def move_file_to_dir(src_file, dest_dir):
    shutil.move(src_file, dest_dir)
    return os.pathsep.join(dest_dir, os.path.basename(src_file))


def create_empty_file(path):
    os.makedirs(os.path.dirname(path))
    f = open("path", "w")
    f.close()


def __base_name_parts(fpath):
    name = os.path.basename(fpath)
    main, *ext = name.rsplit(".", 1)
    if ext:
        return main, ext[0]
    else:
        return main, ""

def get_extension(fpath):
    return __base_name_parts(fpath)[1]

def get_nonext_basename(fpath):
    return __base_name_parts(fpath)[0]

def normalize_path(in_name):
    updated = re.sub(r"[\\]+", "/", in_name)
    updated = re.sub(r"[/]+", "/", updated)
    return updated



class MultiElement:

    def __init__(self, elements):
        self.__elements = elements

    def get_instance_count(self):
        return len(self.__elements)

    def get_element_at_index(self, index):
        return self.__elements[index]




class SetuSvcObjectManager:
    TESTSESSION_HANDLERS = {}

    @classmethod
    def register_testsession_handler(cls, handler):
        cls.TESTSESSION_HANDLERS[handler.setu_id] = handler

    @classmethod
    def deregister_testsession_handler(cls, handler):
        #cascading calls to all test handler objects needed.
        del cls.TESTSESSION_HANDLERS[handler.setu_id]

    @classmethod
    def get_testsession_handler(cls, setu_id):
        return cls.TESTSESSION_HANDLERS[setu_id]

    @classmethod
    def has_active_testsession(cls):
        return len(cls.TESTSESSION_HANDLERS) > 0

    @classmethod
    def deregister_all_existing_testsessions(cls):
        testsession_handles = tuple(cls.TESTSESSION_HANDLERS.values())
        for handler in testsession_handles:
            cls.deregister_testsession_handler(handler)




'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import re
import inspect
from decimal import getcontext, Decimal, ROUND_HALF_EVEN

from arjuna.unitee.validation import checks
from .testpoint import *
import math
import re

class Assertion:
    def __init__(self, purpose, subject):
        self.__purpose = purpose
        self.__subject = subject
        self.__offset = "0.00000000000000000001"
        self.__rounding = ROUND_HALF_EVEN
        self._step = None

    def __create_step(self):
        self._step = Step()
        self._step.purpose = self.__purpose
        frame = inspect.stack()[3]
        self._step.source = "{}.py:{}:L{}".format(inspect.getmodule(frame[0]).__name__, frame[3], frame[2])
        self._step.step_type = StepType.AssertStep

    def evaluate(self):
        self._step.evaluate()

    def _configure_step(self):
        self._step.add_to_state()
        self._step.evaluate()

    def is_true(self):
        self.__create_step()
        self._step.expectation = "Should be True"
        try:
            if checks.is_true(self.__subject):
                self._step.observation = "True"
                self._step.assert_message = "Assertion fulfilled."
            else:
                self._step.observation = "False"
                self._step.set_failure()
                self._step.assert_message = "Condition is False."
        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)
        self._configure_step()

    def is_false(self):
        self.__create_step()
        self._step.expectation = "Should be False"
        try:
            if checks.is_true(self.__subject):
                self._step.observation = "True"
                self._step.set_failure()
                self._step.assert_message = "Condition is True."
            else:
                self._step.observation = "False"
                self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()

    def __update_check_for_null_expected_object(step):
        step.set_error()
        step.assert_message = "Expected object can not be None"
        step.observation = "Error: An error occured before this could be evaluated."

    def __set_check_error_for_assertion_issue(self, e):
        self._step.set_error()
        self._step.assert_message = "Error in assertion. {}".format(e)

    def is_equal_to(self, expected):
        if type(self.__subject) is float or type(expected) is float:
            return self.__is_almost_equal_to(expected)
        self.__create_step()
        has_failed = False
        try:
            if checks.is_none(expected):
                if checks.is_none(self.__subject):
                    pass  # None = None
                else:
                    has_failed = True

            type_mismatch = False
            matched = False
            if type(self.__subject) is not type(self.__subject):
                if type(self.__subject) in {'int', 'float'} and type(expected) in {'int', 'float'}:
                    pass
                else:
                    type_mismatch = True
                    has_failed = True
                    # we need to handle this
            if not type_mismatch:
                self._step.expectation = 'Should be equal to Object of type={0} and value=>{1}<'.format(type(expected).__name__, expected)
                self._step.observation = 'Object of type={0} and value=>{1}<'.format(type(self.__subject).__name__, self.__subject)

                has_failed = checks.are_not_equal(self.__subject, expected)

                if has_failed:
                    self._step.set_failure()
                    self._step.assert_message = "Objects are not equal"
                else:
                    self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()

    def is_not_equal_to(self, expected):
        self.__create_step()
        has_failed = False
        try:
            if checks.is_none(expected):
                if checks.is_none(self.__subject):
                    has_failed = False
                else:
                    pass

            type_mismatch = False
            matched = False
            if type(self.__subject) is not type(self.__subject):
                if type(self.__subject) in {'int', 'float'} and type(expected) in {'int', 'float'}:
                    pass
                else:
                    type_mismatch = True
                    has_failed = True
                    # we need to handle this

            if not type_mismatch:
                self._step.expectation = 'Should not be equal to Object of type={0} and value=>{1}<'.format(type(expected).__name__, expected)
                self._step.observation = 'Object of type={0} and value=>{1}<'.format(type(self.__subject).__name__, self.__subject)

                has_failed = checks.are_equal(self.__subject, expected)

                if has_failed:
                    self._step.set_failure()
                    self._step.assert_message = "Objects are equal"
                else:
                    self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()

    def rounded_as(self, strategy):
        self.__rounding = strategy
        return self

    def with_max_offset(self, offset):
        if offset is None:
            return
        if type(offset) in {int, float}:
            f_offset = Decimal(offset).quantize(Decimal("0.0001"))
            if f_offset <= 0 or f_offset >= 1:
                raise Exception("Max offset provided as float should be a non-negative value such that 0 < offset < 1 after rounding to 4 decimal places.")
            self.__offset = re.sub(r"[0]+$", "", str(f_offset))
        elif type(offset) is str:
            try:
                f_offset = float(offset)
                if f_offset <= 0 or f_offset >= 1:
                    raise Exception(
                        "Max offset provided as string should be a non-negative value such that 0 < offset < 1.")
                else:
                    self.__offset = offset
            except:
                raise Exception("Max offset string must contain a floating number.")
        else:
            raise Exception("Max offset can be a float or string")
        return self

    def __is_almost_equal_to(self, expected):
        self.__create_step()
        has_failed = False
        try:

            not_allowed = False
            lhs = self.__subject
            rhs = expected
            allowed_type_set = {int, float}
            if type(expected) in allowed_type_set and type(self.__subject) in allowed_type_set:
                zero, places = self.__offset.split('.', 1)
                precision = len(places)
                allowed_diff = Decimal(self.__offset)
                try:
                    lhs = Decimal(lhs).quantize(allowed_diff, rounding=self.__rounding)
                except Exception as e:
                    print(e)
                    import traceback
                    traceback.print_exc()
                    # No rounding or offsetting needed as decimal places are less than offset precision
                    lhs = Decimal(lhs)
                try:
                    rhs = Decimal(rhs).quantize(allowed_diff, rounding=self.__rounding)
                except Exception as f:
                    print(f)
                    rhs = Decimal(rhs)

                diff = abs(lhs-rhs)
                has_failed = diff > allowed_diff
            else:
                not_allowed = True

            self._step.expectation = 'Should be equal to approxmiately >{}<.{}{}'.format(
                expected,
                ' Max Offset specified: {}.'.format(allowed_diff),
                " Round as per offset with {} strategy.".format(self.__rounding)
            )

            self._step.observation = '>{}<'.format(self.__subject)

            if not_allowed:
                self._step.set_failure()
                self._step.assert_message = ">{0}<[type:{1}] can not be approxmiately compared to >{2}<[type:{3}]".format(
                    self.__subject,
                    type(self.__subject).__name__,
                    expected,
                    type(expected).__name__
                )
            elif has_failed:
                self._step.set_failure()
                self._step.assert_message = "Numbers are not equal as per approximation guidelines."
            else:
                self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            import traceback
            traceback.print_exc()
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()

    def is_same_as(self, expected):
        self.__create_step()
        has_failed = False
        try:
            if checks.is_none(expected):
                if checks.is_none(self.__subject):
                    pass
                else:
                    has_failed = True

            else:
                self._step.expectation = "Should be same as Object of type >{0}< and id >{1}<".format(type(expected).__name__, id(expected))
                self._step.observation = "Object of type >{0}< and id >{1}<".format(type(self.__subject).__name__, id(self.__subject))

                has_failed = checks.are_not_same(self.__subject, expected)

                if has_failed:
                    self._step.set_failure()
                    self._step.assert_message = "Objects are not same."
                else:
                    self._step.assert_message = "Assertion fulfilled."


        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()


    def is_not_same_as(self, expected):
        self.__create_step()
        has_failed = False
        try:
            if checks.is_none(expected):
                if checks.is_none(self.__subject):
                    has_failed = True
                else:
                    pass
            else:
                self._step.expectation = "Should not be same as Object of type >{0}< and id >{1}<".format(type(expected).__name__, id(expected))
                self._step.observation = "Object of type >{0}< and id >{1}<".format(type(self.__subject).__name__,
                                                                                  id(self.__subject))

                has_failed = checks.are_same(self.__subject, expected)

                if has_failed:
                    self._step.set_failure()
                    self._step.assert_message = "Objects are same."
                else:
                    self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()

    is_different_from = is_not_same_as


    def is_none(self):
        self.__create_step()
        self._step.expectation = "Should be None"
        try:
            if checks.is_not_none(self.__subject):
                self._step.set_failure()
                self._step.observation = "Object of type {0} with id = {1}".format(type(self.__subject).__name__, id(self.__subject))
                self._step.assert_message = "Object is not None"
            else:
                self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()


    def is_not_none(self):
        self.__create_step()
        self._step.expectation = "Should NOT be None"
        try:
            if checks.is_none(self.__subject):
                self._step.set_failure()
                self._step.observation = "None"
                self._step.assert_message = "Supplied object is None."
            else:
                self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()

    def includes(self, content):
        self.__create_step()
        self._step.expectation = "Any iterable containing {}".format(content)
        try:
            has_failed = False
            if checks.is_none(self.__subject):
                has_failed = True
                self._step.observation = "None"
                self._step.assert_message = "Container is None."
            elif type(self.__subject) not in {list, tuple, set, dict}:
                has_failed = True
                self._step.assert_message = "Container is not of type list, tuple, set or dict."
            else:
                self._step.observation = "Object of type={0} and contents=<{1}>".format(type(self.__subject).__name__,
                                                                                        str(self.__subject))
                if type(content) in {list, tuple, set}:
                    for item in content:
                        if item not in self.__subject:
                            has_failed = True
                            self._step.assert_message = "One or more items not included in object."
                            break
                elif type(content) is dict:
                    if type(self.__subject) is dict:
                        for k,v in content.items():
                            if k not in self.__subject or self.__subject[k] != v:
                                has_failed = True
                                self._step.assert_message = "One or more items not included in object."
                                break
                    else:
                        has_failed = True
                        self._step.assert_message = "Container is not of type dict. Attempt to find dict in non-dict failed."
                else:
                    if content not in self.__subject:
                        has_failed = True
                        self._step.assert_message = "Content not included in object."
            if has_failed:
                self._step.set_failure()
            else:
                self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()

    def excludes(self, content):
        self.__create_step()
        self._step.expectation = "Any iterable NOT containing {}".format(content)
        try:
            has_failed = False
            if checks.is_none(self.__subject):
                has_failed = True
                self._step.observation = "None"
                self._step.assert_message = "Container is None."
            elif type(self.__subject) not in {list, tuple, set, dict}:
                has_failed = True
                self._step.assert_message = "Container is not of type list, tuple, set or dict."
            else:
                self._step.observation = "Object of type={0} and contents=<{1}>".format(type(self.__subject).__name__,
                                                                                        str(self.__subject))
                if type(content) in {list, tuple, set}:
                    for item in content:
                        if item in self.__subject:
                            has_failed = True
                            self._step.assert_message = "One or more items included in object."
                            break
                elif type(content) is dict:
                    if type(self.__subject) is dict:
                        for k,v in content.items():
                            if k in self.__subject and self.__subject[k] == v:
                                has_failed = True
                                self._step.assert_message = "One or more items included in object."
                                break
                    else:
                        has_failed = True
                        self._step.assert_message = "Container is not of type dict. Attempt to find dict in non-dict failed."
                else:
                    if content in self.__subject:
                        has_failed = True
                        self._step.assert_message = "Content included in object."
            if has_failed:
                self._step.set_failure()
            else:
                self._step.assert_message = "Assertion fulfilled."

        except Exception as e:
            self.__set_check_error_for_assertion_issue(e)

        self._configure_step()

    def execute(self):
        pass

class LinkedAssertion(Assertion):
    def __init__(self, stepper, subject):
        super().__init__(stepper.purpose, subject)
        self._stepper = stepper

    def _configure_step(self):
        if not self._stepper._is_soft_asserter():
            super()._configure_step()
        else:
            self._step.add_to_state()
            self._stepper._add_step(self._step)


class PendingAssertion(Assertion):
    def __init__(self, stepper, subject):
        super().__init__(stepper.purpose, subject)
        self._stepper = stepper
        self.__pending_func = None
        self.__pending_args = None

    def __assign_caller(self, func, args):
        self.__pending_func = func
        self.__pending_args = args
        self._stepper._add_assertion(self)

    def is_true(self, *vargs):
        self.__assign_caller(super().is_true, vargs)

    def is_false(self, *vargs):
        self.__assign_caller(super().is_false, vargs)

    def is_equal_to(self, *vargs):
        self.__assign_caller(super().is_equal_to, vargs)

    def is_not_equal_to(self, *vargs):
        self.__assign_caller(super().is_not_equal_to, vargs)

    def is_same_as(self, *vargs):
        self.__assign_caller(super().is_same_as, vargs)

    def is_not_same_as(self, *vargs):
        self.__assign_caller(super().is_not_same_as, vargs)

    is_different_from = is_not_same_as

    def is_none(self, *vargs):
        self.__assign_caller(super().is_none, vargs)

    def is_not_none(self, *vargs):
        self.__assign_caller(super().is_not_none, vargs)

    def includes(self, *vargs):
        self.__assign_caller(super().includes, vargs)

    def excludes(self, *vargs):
        self.__assign_caller(super().excludes, vargs)

    def execute(self):
        if self.__pending_args:
            self.__pending_func(*self.__pending_args)
        else:
            self.__pending_func()

    def _configure_step(self):
        if not self._stepper._is_soft_asserter():
            super()._configure_step()
        else:
            self._step.add_to_state()
            self._stepper._add_step(self._step)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import sys
import os
import argparse
from functools import partial

from arjuna.unitee.enums import *

from .validation import *


class Parser:

    def __init__(self):
        self.parser = None

    def get_parser(self):
        return self.parser


class SetuParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-p', '--port', dest="setu.port", default=9000, type=port, help='Setu network port')

    def process(self, arg_dict):
        pass


class ProjectParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        #self.parser.add_argument('-w', '--workspace-dir', dest="workspace_dir", type=str, help='Workspace Directory')
        self.parser.add_argument('-p', '--project-root-dir', dest="project.root.dir", type=project_dir, help = 'Valid absolute Project root directory. Name of project (Alnum 3-30 length. Only lower case letters.).')

    def process(self, arg_dict):
        # arg_dict['workspace.dir'] = arg_dict['workspace_dir']
        # del arg_dict['workspace_dir']
        if arg_dict['project.root.dir'] is None:
            print("Fatal Error in CLI processing. You must provide a valid project root directory using -p or --project-dir switch", file=sys.stderr)
            sys.exit(1)


class NewProjectParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        #self.parser.add_argument('-w', '--workspace-dir', dest="workspace_dir", type=str, help='Workspace Directory')
        self.parser.add_argument('-p', '--project-root-dir', dest="project.root.dir", type=new_project_dir, help = 'Absolute non-existing Project root directory. Name of project (Alnum 3-30 length. Only lower case letters.).')
        self.parser.add_argument('--non-unitee', dest="project.is_not_unitee", action="store_true", help = 'Pass this switch for creating a non-UniTEE project.')

    def process(self, arg_dict):
        # arg_dict['workspace.dir'] = arg_dict['workspace_dir']
        # del arg_dict['workspace_dir']
        if arg_dict['project.root.dir'] is None:
            print("Fatal Error in CLI processing. You must provide a valid project root directory using -p or --project-dir switch", file=sys.stderr)
            sys.exit(1)


class RunParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument("-rid", "--runid", nargs=1, dest="runid", type=partial(lname_check, "Run ID"), help = 'Alnum 3-30 length. Only lower case letters.')
        self.parser.add_argument('-ar', '--active-reporters', dest="unitee.project.active.reporters",
                                 metavar=('R1','R2'), nargs='+',
                                 type=ustr,
                                 choices=[i for i in ActiveReporterNames.__members__],
                                 help='One or more valid active state names: ' + str([i for i in ActiveReporterNames.__members__]))
        self.parser.add_argument('-dr', '--deferred-reporters', dest="unitee.project.deferred.reporters",
                                 metavar=('R1','R2'), nargs='+',
                                 type=ustr,
                                 choices=[i for i in DeferredReporterNames.__members__],
                                 help='One or more valid deferred state names: ' + str([i for i in DeferredReporterNames.__members__]))

        self.parser.add_argument('-aco', '--arjuna-central-option', dest="aco",
                                 nargs=2,
                                 action='append',
                                 metavar=('option', 'value'),
                                 help='Arjuna Central Option. Can pass any number of these switches as -aco x y -aco z t')

        self.parser.add_argument('-ato', '--arjuna-test-option', dest="ato",
                                 nargs=2,
                                 action='append',
                                 metavar=('option', 'value'),
                                 help='Arjuna Test Option. Can pass any number of these switches as -aco x y -ato z t')

        self.parser.add_argument('-uco', '--user-central-option', dest="uco",
                                 nargs=2,
                                 action='append',
                                 metavar=('option', 'value'),
                                 help='User Central Option. Can pass any number of these switches as -aco x y -uco z t')

        self.parser.add_argument('-uto', '--user-text-option', dest="uto",
                                 nargs=2,
                                 action='append',
                                 metavar=('option', 'value'),
                                 help='User Test Option. Can pass any number of these switches as -aco x y -uto z t')

    def process(self, arg_dict):
        pass

class SessionParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-s', '--session-name', dest="unitee.project.session.name", type=partial(lname_check, "Session"), help='Existing session template name.')

    def process(self, arg_dict):
        if arg_dict['unitee.project.session.name'] is None:
            print("Fatal Error in CLI processing. You must provide a valid session name using -s or --session-name switch", file=sys.stderr)
            sys.exit(1)


class GroupParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-g', '--group-name', dest="group.name", type=partial(lname_check, "Group"), help='Existing group template name.')

    def process(self, arg_dict):
        pass


class NamesParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-cm', '--cmodules', dest="cmodules", metavar=('M1','M2'), default=None, nargs='+', help='One or more names/patterns for considering test modules.')
        self.parser.add_argument('-im', '--imodules', dest="imodules", metavar=('M1','M2'), default=None, nargs='+',
                         help='One or more names/patterns for ignoring test modules.')
        self.parser.add_argument('-cf', '--cfunctions', dest="cfunctions", metavar=('F1','F2'), default=None, nargs='+', help='One or more names/patterns for considering test functions.')
        self.parser.add_argument('-if', '--ifunctions', dest="ifunctions", metavar=('F1','F2'), default=None, nargs='+',
                         help='One or more names/patterns for ignoring test functions.')

    def process(self, arg_dict):
        pass


import logging
from arjuna.setu.testsession.impl import TestSession
from arjuna.setuext.guiauto.adapter.automator import GuiAutomatorHandler
from arjuna.tpi.enums import ArjunaOption
from arjuna.setu.dispatcher.testsession import TestSessionDispatcher
from arjuna.setuext.guiauto.adapter.gui import GuiHandlerManager
from arjuna.setu import Setu
from arjuna.setu.adapter import Handler


class TestSessionHandler(Handler):

    def __init__(self):
        self.__testsession = None
        self.__automator_handlers = {}
        self.__conf_handler = None
        self.__databroker_handler = None
        self.__dispatcher = TestSessionDispatcher()
        self.__project_config_loaded = False
        self.__guimgr = None

    def register_gui_automator_handler(self, handler):
        self.__automator_handlers[handler.setu_id] = handler

    def deregister_gui_automator_handler(self, handler):
        del self.__automator_handlers[handler.setu_id]

    def __get_gui_automator_handler(self, setu_id):
        return self.__automator_handlers[setu_id]

    @property
    def setu_id(self):
        return self.__testsession.setu_id

    def init(self, root_dir, cliConfig=None):
        self.__testsession = TestSession()
        self.__testsession.init(root_dir, cliConfig)
        config = self.__testsession.configurator.create_project_conf()
        Setu.init_logger(self.__testsession.setu_id, config.setu_config.value(ArjunaOption.LOG_DIR))
        self.__conf_handler = TestSessionConfHandler(self.__testsession.configurator)
        self.__databroker_handler = TestSessionDataBrokerHandler(self, self.__testsession.data_broker)
        self.__guimgr = GuiHandlerManager(config)
        return config

    @property
    def conf_handler(self):
        return self.__conf_handler

    @property
    def datasource_handler(self):
        return self.__databroker_handler

    @property
    def dispatcher(self):
        return self.__dispatcher

    @property
    def gui_manager(self):
        return self.__guimgr

    def get_automator_handler(self, json_dict):
        return self.__get_gui_automator_handler(self._pop_arg(json_dict, "automatorSetuId"))

    def get_gui_handler(self, json_dict):
        return self.__guimgr.get_gui_handler(self._pop_arg(json_dict, "guiSetuId"))

    def take_session_action(self, action, json_dict):
        return getattr(self, action)(json_dict)

class TestSessionConfHandler(Handler):

    def __init__(self, configurator):
        self.__configurator = configurator

    @property
    def configurator(self):
        return self.__configurator

    def take_action(self, action, json_args):
        return getattr(self, action.lower())(**json_args)

    def register_new_config(self, arjunaOptions, userOptions=None, hasParent=False, parentConfigId=None):
        # Registering a config is post project conf registration. If no project conf, set it to true.
        self.__project_config_loaded = True
        config = self.__configurator.register_config(arjunaOptions, hasParent, userOptions, parentConfigId)
        out_map = {'configSetuId' : config.setu_id}
        out_map.update(config.as_json_dict())
        return out_map

    def get_arjuna_option_value(self, configSetuId, option):
        return {"value": self.__configurator.get_arjuna_option_value(configSetuId, option)}

    def get_user_option_value(self, configSetuId, option):
        return {"value": self.__configurator.get_user_option_value(configSetuId, option)}


class TestSessionDataBrokerHandler(Handler):

    def __init__(self, testsession_handler, data_broker):
        self.__data_broker = data_broker
        self.__testsession_handler = testsession_handler

    def take_action(self, action_type, json_dict):
        return getattr(self, action_type.name.lower())(**json_dict) 

    def create_file_data_source(self, fileName, recordType, **json_dict):
        data_dir = self.__testsession_handler.conf_handler.configurator.get_central_arjuna_option_value(ArjunaOption.DATA_SOURCES_DIR.name)
        return {"dataSourceSetuId" : self.__data_broker.create_file_data_source(data_dir, fileName, recordType, **json_dict)}  

    def get_next_record(self, sourceSetuId):
        try:
            return {"finished" : False, "record" : self.__data_broker.get_next_record(sourceSetuId)}
        except Exception:
            return {"finished" : True}

    def get_all_records(self, sourceSetuId):
        try:
            return {"records" : self.__data_broker.get_all_records(sourceSetuId)}
        except Exception:
            return {"finished" : True}

    def reset(self, sourceSetuId):
        return {"finished" : False, "record" : self.__data_broker.reset(sourceSetuId)}


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import inspect
from decimal import ROUND_HALF_EVEN

from arjuna.lib.enums import *
from .testpoint import *
from arjuna.unitee.enums import *
from arjuna.unitee.reporter.result.types import *
from .assertion import *

class _Asserter:

    def __init__(self, purpose):
        self.__purpose = purpose

    @property
    def purpose(self):
        return self.__purpose

    def evaluate(self):
        pass

    def _is_soft_asserter(self):
        return False

    def assert_that(self, subject):
        assertion = LinkedAssertion(self, subject)
        return assertion

class _SoftAsserter(_Asserter):

    def __init__(self, purpose):
        super().__init__(purpose)
        self.__steps = []

    def _add_step(self, step):
        self.__steps.append(step)

    def _is_soft_asserter(self):
        return True

    def evaluate(self):
        for step in self.__steps:
            step.evaluate()

class _LazyAsserter(_SoftAsserter):

    def __init__(self, purpose):
        super().__init__(purpose)
        self.__assertions = []

    def execute(self):
        for assertion in self.__assertions:
            assertion.execute()

    def _add_assertion(self, assertion):
        self.__assertions.append(assertion)

    def assert_that(self, subject):
        assertion = PendingAssertion(self, subject)
        return assertion

class Steps:

    @staticmethod
    def __step(src, step_type, step_result_type, purpose, expectation, observation, assert_message, **kwargs):
        step = Step()
        step.step_type = step_type
        step.purpose = purpose
        step.relation = "-"
        step.step_props = kwargs
        step.source = src

        if expectation is not None and observation is not None:
            step.expectation = expectation
            step.observation = observation
        elif expectation is None and observation is None:
            pass
        else:
            raise Exception("Expectation and Actual observation should be provided or ignored together.")

        if step_result_type == ResultTypeEnum.NOTHING:
            step.rtype = "-"
        elif step_result_type == ResultTypeEnum.FAIL:
            step.set_failure()
        elif step_result_type == ResultTypeEnum.ERROR:
            step.set_error()

        step.assert_message = assert_message and assert_message or "-"

        step.add_to_state()
        step.evaluate()

    # Plan to include take_screenshot=False as a keyword argument for different steps.
    # Also plan to have a screenshot step.
    # Also plan to capture the exception trace in the step and issue info.

    @staticmethod
    def passed(purpose, *, expectation=None, observation=None,  **kwargs):
        src = Steps.__get_caller()
        Steps.__step(src, StepType.ExecStep, ResultTypeEnum.PASS, purpose, expectation, observation, None, **kwargs)

    @staticmethod
    def failed(purpose, *, expectation=None, observation=None, message=None, **kwargs):
        src = Steps.__get_caller()
        Steps.__step(src, StepType.ExecStep, ResultTypeEnum.FAIL, purpose, expectation, observation, message, **kwargs)

    @staticmethod
    def erred(purpose, message=None, **kwargs):
        src = Steps.__get_caller()
        Steps.__step(src, StepType.ExecStep, ResultTypeEnum.ERROR, purpose, None, None, message, **kwargs)

    @staticmethod
    def log(purpose, **kwargs):
        from arjuna.tpi import Arjuna
        logger = Arjuna.get_logger()
        logger.debug(purpose)
        src = Steps.__get_caller()
        Steps.__step(src, StepType.LogStep, ResultTypeEnum.NOTHING, purpose, None, None, None, **kwargs)

    @staticmethod
    def __get_caller():
        frame = inspect.stack()[2]
        return "{}.py:{}:L{}".format(inspect.getmodule(frame[0]).__name__, frame[3], frame[2])

    @staticmethod
    def validate(purpose):
        return _Asserter(purpose)

    @staticmethod
    def soft_validate(purpose):
        return _SoftAsserter(purpose)

    @staticmethod
    def lazy_validate(purpose):
        return _LazyAsserter(purpose)

    @staticmethod
    def assert_true(purpose, actual):
        assertion = Assertion(purpose, actual)
        assertion.is_true()

    @staticmethod
    def assert_false(purpose, actual):
        assertion = Assertion(purpose, actual)
        assertion.is_false()

    @staticmethod
    def assert_equal(purpose, actual, expected):
        assertion = Assertion(purpose, actual)
        assertion.is_equal_to(expected)

    @staticmethod
    def assert_not_equal(purpose, actual, expected):
        assertion = Assertion(purpose, actual)
        assertion.is_not_equal_to(expected)

    @staticmethod
    def assert_almost_equal(purpose, actual, expected, max_offset=None, rounded_as=ROUND_HALF_EVEN):
        assertion = Assertion(purpose, actual)
        assertion.with_max_offset(max_offset)
        assertion.rounded_as(rounded_as)
        assertion.is_equal_to(expected)

    @staticmethod
    def assert_same(purpose, actual, expected):
        assertion = Assertion(purpose, actual)
        assertion.is_same_as(expected)

    @staticmethod
    def assert_not_same(purpose, actual, expected):
        assertion = Assertion(purpose, actual)
        assertion.is_not_same_as(expected)

    assert_different = assert_not_same

    @staticmethod
    def assert_none(purpose, actual):
        assertion = Assertion(purpose, actual)
        assertion.is_none()

    @staticmethod
    def assert_not_none(purpose, actual):
        assertion = Assertion(purpose, actual)
        assertion.is_not_none()

    @staticmethod
    def assert_includes(purpose, container, content):
        assertion = Assertion(purpose, container)
        assertion.includes(content)

    @staticmethod
    def assert_excludes(purpose, container, content):
        assertion = Assertion(purpose, container)
        assertion.excludes(content)


from arjuna.setu.types import SetuManagedObject
from arjuna.setuext.guiauto.impl.element.guielement import GuiElement
from arjuna.setuext.guiauto.impl.locator.emd import SimpleGuiElementMetaData

from arjuna.setu import Setu

class FrameContainer(SetuManagedObject):
    def __init__(self, automator):
        super().__init__()
        self.__automator = automator

    @property
    def automator(self):
        return self.__automator

    def _act(self, json_dict):
        return self.__automator.actor_callable(json_dict)

    def __check_tag(self, wrapped_element):
        tag = wrapped_element.get_tag_name()
        if tag.lower() != "iframe":
            raise Exception("The element should have a 'iframe' tag for IFrame element. Found: " + tag)

    def create_frame(self, locator_meta_data):
        found = False
        frame = None
        for locator in locator_meta_data.locators: 
            try:
                if locator.ltype.name == "INDEX":
                    index = locator.lvalue
                    emd = SimpleGuiElementMetaData("xpath", "//iframe")
                    multi_element = self.automator.create_multielement(emd)
                    multi_element.find()
                    wrapped_element = multi_element.get_instance_at_index(index)
                    self.__check_tag(wrapped_element)
                    frame = IPartialFrame(self.automator, self, multi_element, wrapped_element)
                else:
                    emd = SimpleGuiElementMetaData(locator.ltype.name, locator.lvalue)
                    wrapped_element = self.automator.create_element(emd)
                    self.__check_tag(wrapped_element)
                    frame = IFrame(self.automator, self, wrapped_element)

                found = True
            except Exception as e:
                print(e)
                continue

        if not found:
            raise Exception("Could not locate frame with locator(s): {}".format(locator_meta_data.locators))
        else:
            self.automator.add_frame(frame)
            return frame

class DomRoot(FrameContainer):

    def __init__(self, automator):
        super().__init__(automator)
        self.__frame_context = "root"
        self.automator.add_frame(self)

    @property
    def frame_context(self):
        return self.__frame_context

    def __set_frame_context_str(self, name):
        self.__frame_context = name
        Setu.get_logger().debug("Automator is in {} frame".format(self.__frame_context))

    def set_frame_context(self, frame):
        self.__set_frame_context_str(frame.get_setu_id())

    def set_frame_context_as_root(self):
        self.__set_frame_context_str("root")  

    def focus(self):
        self.automator.dispatcher.focus_on_dom_root()
        self.set_frame_context_as_root() 

# UUID is for client reference. Agent does not know about this.
class IFrame(FrameContainer):

    def __init__(self, automator, dom_root, wrapped_element):
        super().__init__(automator)
        self.__dom_root = dom_root
        self.__parent_frames = []
        self.__wrapped_element = wrapped_element

    @property
    def dom_root(self):
        return self.__dom_root

    @property
    def wrapped_element(self):
        return self.__wrapped_element

    def set_parents(self, parents):
        self.__parent_frames = parents

    def _act(self, json_dict):
        return self.dom_root._act(json_dict)

    def _focus_on_parents(self):
        if self.__parent_frames:
            for parent in self.__parent_frames:
                parent.focus()

    def focus(self):
        self.wrapped_element.find()
        self.automator.dispatcher.focus_on_frame(self.wrapped_element.setu_id)
        self.dom_root.set_frame_context(self)

    # def focus_on_parent(self):
    #     self._act(TestAutomatorActionBodyCreator.jump_to_parent_frame())
    #     if self.__parent_frames:
    #         self.dom_root.set_frame_context(self.__parent_frames[-1])
    #     else:
    #         self.dom_root.set_frame_context_as_root()

    def get_parent(self):
        if self.__parent_frames:
            return self.__parent_frames[-1]
        else:
            return self.dom_root


class IPartialFrame(IFrame):

    def __init__(self, automator, dom_root, melement, wrapped_element):
        super().__init__(automator, dom_root, wrapped_element)
        self.__melement = melement

    def focus(self):
        self.__melement.find()
        self.automator.dispatcher.focus_on_frame(self.__melement.setu_id, True, self.wrapped_element._get_instance_number())
        self.dom_root.set_frame_context(self)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os
import abc
import threading

from arjuna.lib.utils import file_utils
from arjuna.unitee.enums import *
from arjuna.lib.thread.decorators import *
from arjuna.unitee.reporter.mcr import *
from arjuna.unitee.reporter.cr import *
from arjuna.unitee.reporter.excel import *

from arjuna.tpi.enums import ArjunaOption

class __Reporter(metaclass=abc.ABCMeta):
    def __init__(self):
        self.__reporters = []

    @property
    def reporters(self):
        return self.__reporters

    def _add_reporter(self, reporter):
        self.reporters.append(reporter)

    def set_up(self):
        for reporter in self.reporters:
            reporter.set_up()

    def tear_down(self):
        for reporter in self.reporters:
            reporter.tear_down()

class NonActiveReporter(__Reporter, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def generate(self):
        pass

class OfflineReporter(NonActiveReporter):
    def __init__(self):
        super().__init__()

    def generate(self):
        pass

class DeferredReporter(NonActiveReporter):
    def __init__(self, rr_dir):
        super().__init__()
        self.__rr_dir = rr_dir
        deferred_formats = Arjuna.get_central_config().get_arjuna_option_value(ArjunaOption.UNITEE_PROJECT_DEFERRED_REPORTERS).as_enum_list(DeferredReporterNames)
        for rfmt in deferred_formats:
            if rfmt == DeferredReporterNames.EXCEL:
                rdir = os.path.join(self.__rr_dir, "/excel")
                os.makedirs(rdir)
                self._add_reporter(ExcelReporter(rdir))
            else:
                raise Exception("{} deffered reporting not implemented yet.".format(rfmt))

    def generate(self):
        pass

class ActiveReporter(__Reporter):

    def __init__(self):
        super().__init__()
        self.lock = threading.RLock()
        rdir = Arjuna.get_central_config().get_arjuna_option_value(ArjunaOption.REPORT_DIR).as_string()
        file_utils.delete_dir_if_exists(rdir)
        self.__rr_dir = Arjuna.get_central_config().get_arjuna_option_value(ArjunaOption.UNITEE_PROJECT_RUN_REPORT_DIR).as_string()
        os.makedirs(self.__rr_dir)

    def set_up(self):
        active_formats = Arjuna.get_central_config().get_arjuna_option_value(
            ArjunaOption.UNITEE_PROJECT_ACTIVE_REPORTERS).as_enum_list(ActiveReporterNames)
        for afmt in active_formats:
            if afmt == ActiveReporterNames.MIN_CONSOLE:
                self._add_reporter(MinimalConsoleReporter())
            elif afmt == ActiveReporterNames.CONSOLE:
                self._add_reporter(ConsoleReporter())
            elif afmt == ActiveReporterNames.EXCEL:
                self._add_reporter(ExcelReporter())
            else:
                raise Exception("{} active reporting not implemented yet.".format(afmt))
        super().set_up()
        for reporter in self.reporters:
            reporter.update_info_headers([i.name for i in InfoEntryEnum])
            reporter.update_execution_headers([i.name for i in ResultEntryEnum])
            reporter.update_step_headers([i.name for i in StepEntryEnum])
            reporter.update_issue_headers([i.name for i in IssueEntryEnum])

    @sync_method('lock')
    def update_info(self, reportable):
        for reporter in self.reporters:
            reporter.update_info(reportable.get_dict())

    @sync_method('lock')
    def update_result(self, reportable):
        if reportable.has_issues():
            for issue in reportable.issues:
                for reporter in self.reporters:
                    reporter.update_issue(issue.get_dict())
        for reporter in self.reporters:
            reporter.update_test_object_result(reportable.get_reportable())

    def report(self, offline=False):
        dr = DeferredReporter(self.__rr_dir)
        dr.set_up()
        dr.generate()
        dr.tear_down()

    def archive(self):
        import shutil
        shutil.copytree(self.__rr_dir,
                        os.path.join(
                            ArjunaCore.config.value(UniteePropertyEnum.ARCHIVES_DIR),
                            ArjunaCore.config.value(UniteePropertyEnum.IRUNID)
                        ))


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import subprocess


def execute_windows_process(proc_name, arg_list=[]):
    '''
    Executes a windows process using the command shell in termination mode.

    Arguments:
    proc_name: Name of process
    arg_list: (Optional) A list of strings that are to be passed as command line arguments to said process.
    '''
    p = subprocess.Popen(
        ["cmd.exe", "/C", proc_name] + arg_list,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    stdout, stderr = p.communicate()

    return p.returncode, stdout, stderr


def kill_process(process_name):
    returncode, stdout, stderr = execute_windows_process("taskkill", ["/F", "/IM", process_name])
    if returncode != 0:
        raise Exception("Not able to kill process: " + process_name)


def start_service(name):
    returncode, stdout, stderr = execute_windows_process("sc", ["start", '"%s"'.format(name)])
    if returncode != 0 or not is_service_running(name):
        raise Exception("Unable to start service: " + name)


def stop_service(name):
    returncode, stdout, stderr = execute_windows_process("sc", ["stop", '"%s"'.format(name)])
    if returncode != 0 or is_service_running(name):
        raise Exception("Unable to stop service: " + name)


def is_service_running(name):
    returncode, stdout, stderr = execute_windows_process("sc", ["query", '"%s"'.format(name)])
    if returncode != 0:
        raise Exception("Not able to query service: " + name)
    else:
        return "RUNNING" in stdout



from enum import Enum, auto


class _WithType(Enum):
    ID = auto()
    NAME = auto()
    CLASS_NAME = auto()
    LINK_TEXT = auto()
    XPATH = auto()
    CSS_SELECTOR = auto()

    INDEX = auto()
    CHILD_LOCATOR = auto()
    ASSIGNED_NAME = auto()


class With:

    def __init__(self, with_type, with_value):
        '''

        :param with_type: an enum constant of type WithType
        :param with_value: a string or int or a With object.
        '''
        self.__with_type = with_type
        self.__with_value = None
        self.__is_child_locator = False
        self.__child = None

        if with_type == _WithType.CHILD_LOCATOR:
            if not isinstance(with_value, With):
                raise Exception("For identification with child locator, the argument must be a With object.")
            else:
                self.__is_child_locator = True
                self.__child = with_value
        else:
            self.__with_value = with_value

    def as_map(self):
        map = dict()
        map["withType"] = self.__with_type.name
        if not self.__is_child_locator:
            map["withValue"] = self.__with_value
        else:
            map["withValue"] = self.__child.as_map()
        return map

    @staticmethod
    def id(id):
        return With(_WithType.ID, id)

    @staticmethod
    def name(name):
        return With(_WithType.NAME, name)

    @staticmethod
    def class_name(name):
        return With(_WithType.CLASS_NAME, name)

    @staticmethod
    def link_text(text):
        return With(_WithType.LINK_TEXT, text)

    @staticmethod
    def css_selector(selector):
        return With(_WithType.CSS_SELECTOR, selector)

    @staticmethod
    def xpath(xpath):
        return With(_WithType.XPATH, xpath)

    @staticmethod
    def index(index):
        return With(_WithType.INDEX, index)

    @staticmethod
    def child_locator(with_obj):
        return With(_WithType.CHILD_LOCATOR, with_obj)

    @staticmethod
    def assigned_name(name):
        return With(_WithType.ASSIGNED_NAME, name)


from arjuna.client.guiauto.gui import BaseGui


class DefaultGui(BaseGui):

    def __init__(self, automator, label, def_file_name, parent=None):
        super().__init__(automator, label=label, def_file_name=def_file_name, parent=parent)


class SimpleBaseGui(BaseGui):

    def __init__(self, automator, app_def_dir=""):
        super().__init__(automator, register=False)
        label = self.__class__.__name__
        self.set_label(label)
        self.set_def_file_name("{}/{}.gns".format(app_def_dir, label))
        self._register()


class SimpleBaseChildGui(BaseGui):

    def __init__(self, automator, parent, app_def_dir=""):
        super().__init__(automator, parent=parent, register=False)
        label = self.__class__.__name__
        self.set_label(label)
        self.set_def_file_name("{}/{}.gns".format(app_def_dir, label))
        self._register()


from arjuna.client.core.action import *


HANDLER_NAME_MAP = {
    1 : ("take_direct_action", ""),
    2 : ("take_browser_action", "BROWSER_"),
    3 : ("take_alert_action", "ALERT_"),
    4 : ("take_element_action", "ELEMENT_"),
    5 : ("take_dropdown_action", "DROPDOWN_"),
    6 : ("take_radiogroup_action", "RADIOGROUP_"),
    7 : ("take_domroot_action", "DOMROOT_"),
    8 : ("take_frame_action", "FRAME_"),
    9 : ("take_window_action", "WINDOW_"),
    10 : ("take_main_window_action", "MAIN_WINDOW_"),
    11 : ("take_child_window_action", "CHILD_WINDOW_"),
    12 : ("take_multielement_action", "MULTIELEMENT_")
}

HANDLER_MAP = {
    GuiAutoActionType.DEFINE_ELEMENT : 1,
    GuiAutoActionType.DEFINE_MULTIELEMENT : 1,
    GuiAutoActionType.DEFINE_DROPDOWN : 1,
    GuiAutoActionType.DEFINE_RADIOGROUP: 1,
    GuiAutoActionType.DEFINE_ALERT: 1,

    GuiAutoActionType.QUIT_AUTOMATOR: 1,
    
    GuiAutoActionType.BROWSER_GO_TO_URL: 2,
    GuiAutoActionType.BROWSER_GO_BACK: 2,
    GuiAutoActionType.BROWSER_GO_FORWARD: 2,
    GuiAutoActionType.BROWSER_REFRESH: 2,
    GuiAutoActionType.BROWSER_EXECUTE_JAVASCRIPT: 2,
    
    GuiAutoActionType.DEFINE_MAIN_WINDOW: 1,
    GuiAutoActionType.SET_SLOMO: 1,

    GuiAutoActionType.ALERT_CONFIRM: 3,
    GuiAutoActionType.ALERT_DISMISS: 3,
    GuiAutoActionType.ALERT_GET_TEXT: 3,
    GuiAutoActionType.ALERT_SEND_TEXT: 3,
    
    GuiAutoActionType.ELEMENT_ENTER_TEXT: 4,
    GuiAutoActionType.ELEMENT_SET_TEXT: 4,
    GuiAutoActionType.ELEMENT_CLEAR_TEXT: 4,

    GuiAutoActionType.ELEMENT_CLICK: 4,
    
    GuiAutoActionType.ELEMENT_IDENTIFY: 4,
    GuiAutoActionType.ELEMENT_WAIT_UNTIL_PRESENT: 4,
    GuiAutoActionType.ELEMENT_WAIT_UNTIL_VISIBLE: 4,
    GuiAutoActionType.ELEMENT_WAIT_UNTIL_CLICKABLE: 4,
    
    GuiAutoActionType.ELEMENT_CHECK: 4,
    GuiAutoActionType.ELEMENT_UNCHECK: 4,

    GuiAutoActionType.ELEMENT_GET_SOURCE: 4,

    GuiAutoActionType.MULTIELEMENT_GET_INSTANCE_COUNT: 12,
    GuiAutoActionType.MULTIELEMENT_GET_RANDOM_INDEX: 12,
    
    GuiAutoActionType.DROPDOWN_HAS_VALUE_SELECTED: 5,
    GuiAutoActionType.DROPDOWN_HAS_INDEX_SELECTED: 5,
    GuiAutoActionType.DROPDOWN_SELECT_BY_VALUE: 5,
    GuiAutoActionType.DROPDOWN_SELECT_BY_INDEX: 5,
    GuiAutoActionType.DROPDOWN_GET_FIRST_SELECTED_OPTION_VALUE : 5,        
    GuiAutoActionType.DROPDOWN_HAS_VISIBLE_TEXT_SELECTED: 5,
    GuiAutoActionType.DROPDOWN_GET_FIRST_SELECTED_OPTION_TEXT: 5,
    GuiAutoActionType.DROPDOWN_SELECT_BY_VISIBLE_TEXT: 5,

    GuiAutoActionType.RADIOGROUP_HAS_VALUE_SELECTED: 6,
    GuiAutoActionType.RADIOGROUP_HAS_INDEX_SELECTED: 6,
    GuiAutoActionType.RADIOGROUP_SELECT_BY_VALUE: 6,
    GuiAutoActionType.RADIOGROUP_SELECT_BY_INDEX: 6,
    GuiAutoActionType.RADIOGROUP_GET_FIRST_SELECTED_OPTION_VALUE : 6,

    GuiAutoActionType.DOMROOT_FOCUS: 7,
    GuiAutoActionType.DOMROOT_CREATE_FRAME: 7,
    
    GuiAutoActionType.FRAME_FOCUS: 8,
    GuiAutoActionType.FRAME_CREATE_FRAME: 8,
    GuiAutoActionType.FRAME_GET_PARENT: 8,
    
    GuiAutoActionType.WINDOW_FOCUS: 9,
    GuiAutoActionType.WINDOW_GET_TITLE: 9,
    
    GuiAutoActionType.MAIN_WINDOW_MAXIMIZE: 10,
    GuiAutoActionType.MAIN_WINDOW_CREATE_CHILD_WINDOW: 10,
    GuiAutoActionType.MAIN_WINDOW_GET_LATEST_CHILD_WINDOW: 10,
    GuiAutoActionType.MAIN_WINDOW_CLOSE_ALL_CHILD_WINDOWS: 10,
    
    GuiAutoActionType.CHILD_WINDOW_CLOSE: 11,
}